#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import os
import sys
import textwrap
import threading
import time
import uuid
from pathlib import Path
from typing import Sequence, Iterable, List, Optional, Tuple, Any

import click

from neptune.new.constants import (
    ASYNC_DIRECTORY,
    NEPTUNE_RUNS_DIRECTORY,
    OFFLINE_DIRECTORY,
    OFFLINE_NAME_PREFIX,
)
from neptune.new.envs import NEPTUNE_SYNC_BATCH_TIMEOUT_ENV, PROJECT_ENV_NAME
from neptune.new.exceptions import (
    CannotSynchronizeOfflineRunsWithoutProject,
    NeptuneConnectionLostException, NeptuneException,
    ProjectNotFound,
    RunNotFound,
)
from neptune.new.internal.backends.api_model import Project, ApiRun
from neptune.new.internal.backends.hosted_neptune_backend import HostedNeptuneBackend
from neptune.new.internal.backends.neptune_backend import NeptuneBackend
from neptune.new.internal.containers.disk_queue import DiskQueue
from neptune.new.internal.credentials import Credentials
from neptune.new.internal.operation import Operation

#######################################################################################################################
# Run and Project utilities
#######################################################################################################################


# Set in CLI entry points block, patched in tests
backend: NeptuneBackend = None

retries_timeout = int(os.getenv(NEPTUNE_SYNC_BATCH_TIMEOUT_ENV, "3600"))


def get_run(run_id: str) -> Optional[ApiRun]:
    try:
        return backend.get_run(run_id)
    except RunNotFound:
        return None
    except NeptuneException as e:
        click.echo('Exception while fetching run {}. Skipping run.'.format(run_id), err=True)
        logging.exception(e)
        return None


project_name_missing_message = (
    'Project name not provided. Could not synchronize offline runs. '
    'To synchronize offline run, specify the project name with the --project flag '
    'or by setting the {} environment variable.'.format(PROJECT_ENV_NAME))


def project_not_found_message(project_name: str) -> str:
    return ('Project {} not found. Could not synchronize offline runs. '.format(project_name) +
            'Please ensure you specified the correct project name with the --project flag ' +
            'or with the {} environment variable, or contact Neptune for support.'.format(PROJECT_ENV_NAME))


def get_project(project_name_flag: Optional[str]) -> Optional[Project]:
    project_name = project_name_flag or os.getenv(PROJECT_ENV_NAME)
    if not project_name:
        click.echo(textwrap.fill(project_name_missing_message), file=sys.stderr)
        return None
    try:
        return backend.get_project(project_name)
    except ProjectNotFound:
        click.echo(textwrap.fill(project_not_found_message(project_name)), file=sys.stderr)
        return None


def get_qualified_name(run: ApiRun) -> str:
    return "{}/{}/{}".format(run.workspace, run.project_name, run.short_id)


def is_valid_uuid(val: Any) -> bool:
    try:
        uuid.UUID(val)
        return True
    except ValueError:
        return False


#######################################################################################################################
# Listing runs to be synchronized
#######################################################################################################################

def is_run_synced(run_path: Path) -> bool:
    return all(is_execution_synced(execution_path) for execution_path in run_path.iterdir())


def is_execution_synced(execution_path: Path) -> bool:
    disk_queue = DiskQueue(execution_path, lambda x: x.to_dict(), Operation.from_dict, threading.RLock())
    return disk_queue.is_empty()


def get_offline_runs_ids(base_path: Path) -> List[str]:
    result = []
    if not (base_path / OFFLINE_DIRECTORY).is_dir():
        return []
    for run_path in (base_path / OFFLINE_DIRECTORY).iterdir():
        result.append(run_path.name)
    return result


def partition_runs(base_path: Path) -> Tuple[List[ApiRun], List[ApiRun], int]:
    synced_runs_ids = []
    unsynced_runs_ids = []
    for run_path in (base_path / ASYNC_DIRECTORY).iterdir():
        run_id = run_path.name
        if is_run_synced(run_path):
            synced_runs_ids.append(run_id)
        else:
            unsynced_runs_ids.append(run_id)
    synced_runs = [run for run in map(get_run, synced_runs_ids)]
    unsynced_runs = [run for run in map(get_run, unsynced_runs_ids)]
    not_found = len([exp for exp in synced_runs + unsynced_runs if not exp or exp.trashed])
    synced_runs = [exp for exp in synced_runs if exp and not exp.trashed]
    unsynced_runs = [exp for exp in unsynced_runs if exp and not exp.trashed]

    return synced_runs, unsynced_runs, not_found


offline_run_explainer = '''
Runs which execute offline are not created on the server and they are not assigned to projects;
instead, they are identified by UUIDs like the ones above.
When synchronizing offline runs, please specify the workspace and project using the "--project"
flag. Alternatively, you can set the environment variable
{} to the target workspace/project. See the examples below.
'''.format(PROJECT_ENV_NAME)


def list_runs(base_path: Path, synced_runs: Sequence[ApiRun],
              unsynced_runs: Sequence[ApiRun], offline_runs_ids: Sequence[str]) \
        -> None:
    if not synced_runs and not unsynced_runs and not offline_runs_ids:
        click.echo('There are no Neptune runs in {}'.format(base_path))
        sys.exit(1)

    if unsynced_runs:
        click.echo('Unsynchronized runs:')
        for run in unsynced_runs:
            click.echo('- {}'.format(get_qualified_name(run)))

    if synced_runs:
        click.echo('Synchronized runs:')
        for run in synced_runs:
            click.echo('- {}'.format(get_qualified_name(run)))

    if offline_runs_ids:
        click.echo('Unsynchronized offline runs:')
        for run_id in offline_runs_ids:
            click.echo('- {}{}'.format(OFFLINE_NAME_PREFIX, run_id))
        click.echo()
        click.echo(textwrap.fill(offline_run_explainer, width=90))

    if not unsynced_runs:
        click.echo()
        click.echo('There are no unsynchronized runs in {}'.format(base_path))

    if not synced_runs:
        click.echo()
        click.echo('There are no synchronized runs in {}'.format(base_path))

    click.echo()
    click.echo('Please run with the `neptune sync --help` to see example commands.')


def synchronization_status(base_path: Path) -> None:
    synced_runs, unsynced_runs, not_found = partition_runs(base_path)
    if not_found > 0:
        click.echo(
            "WARNING: {} runs was skipped because they are in trash or do not exist anymore.".format(not_found),
            sys.stderr)
    offline_runs_ids = get_offline_runs_ids(base_path)
    list_runs(base_path, synced_runs, unsynced_runs, offline_runs_ids)


#######################################################################################################################
# Run synchronization
#######################################################################################################################


def sync_run(run_path: Path, qualified_run_name: str) -> None:
    run_id = run_path.name
    click.echo('Synchronising {}'.format(qualified_run_name))
    for execution_path in run_path.iterdir():
        sync_execution(execution_path, run_id)
    click.echo('Synchronization of run {} completed.'.format(qualified_run_name))


def sync_execution(execution_path: Path, run_id: str) -> None:
    disk_queue = DiskQueue(execution_path, lambda x: x.to_dict(), Operation.from_dict, threading.RLock())
    while True:
        batch, version = disk_queue.get_batch(1000)
        if not batch:
            break

        start_time = time.monotonic()
        while True:
            try:
                backend.execute_operations(run_id, batch)
                break
            except NeptuneConnectionLostException as ex:
                if time.monotonic() - start_time > retries_timeout:
                    raise ex
                click.echo("Experiencing connection interruptions. "
                           "Will try to reestablish communication with Neptune. "
                           f"Internal exception was: {ex.cause.__class__.__name__}",
                           sys.stderr)

        disk_queue.ack(version)


def sync_all_registered_runs(base_path: Path) -> None:
    for run_path in (base_path / ASYNC_DIRECTORY).iterdir():
        if not is_run_synced(run_path):
            run_id = run_path.name
            run = get_run(run_id)
            if run:
                sync_run(run_path, get_qualified_name(run))


def sync_selected_registered_runs(base_path: Path, qualified_runs_names: Sequence[str]) -> None:
    for name in qualified_runs_names:
        run = get_run(name)
        if run:
            run_path = base_path / ASYNC_DIRECTORY / str(run.id)
            if run_path.exists():
                sync_run(run_path, name)
            else:
                click.echo("Warning: Run '{}' does not exist in location {}".format(name, base_path),
                           file=sys.stderr)


def register_offline_run(project: Project) -> Optional[ApiRun]:
    try:
        return backend.create_run(project.id)
    except Exception as e:
        click.echo('Exception occurred while trying to create a run '
                   'on the Neptune server. Please try again later',
                   file=sys.stderr)
        logging.exception(e)
        return None


def move_offline_run(base_path: Path, offline_id: str, server_id: str) -> None:
    (base_path / ASYNC_DIRECTORY / server_id).mkdir(parents=True)
    (base_path / OFFLINE_DIRECTORY / offline_id).rename(base_path / ASYNC_DIRECTORY / server_id / "exec-0-offline")


def register_offline_runs(base_path: Path, project: Project,
                          offline_runs_ids: Iterable[str]) -> List[ApiRun]:
    result = []
    for run_id in offline_runs_ids:
        if (base_path / OFFLINE_DIRECTORY / run_id).is_dir():
            run = register_offline_run(project)
            if run:
                move_offline_run(base_path, offline_id=run_id, server_id=run.id)
                click.echo('Offline run {} registered as {}'
                           .format(run_id, get_qualified_name(run)))
                result.append(run)
        else:
            click.echo('Offline run with UUID {} not found on disk.'.format(run_id), err=True)
    return result


def is_offline_run_name(name: str) -> bool:
    return name.startswith(OFFLINE_NAME_PREFIX) and is_valid_uuid(name[len(OFFLINE_NAME_PREFIX):])


def sync_offline_runs(base_path: Path, project_name: Optional[str], offline_run_ids: Sequence[str]):
    if offline_run_ids:
        project = get_project(project_name)
        if not project:
            raise CannotSynchronizeOfflineRunsWithoutProject
        registered_runs = register_offline_runs(base_path, project, offline_run_ids)
        offline_runs_names = [get_qualified_name(exp) for exp in registered_runs]
        sync_selected_registered_runs(base_path, offline_runs_names)


def sync_selected_runs(base_path: Path, project_name: Optional[str],
                       runs_names: Sequence[str]) -> None:
    other_runs_names = [name for name in runs_names if not is_offline_run_name(name)]
    sync_selected_registered_runs(base_path, other_runs_names)

    offline_runs_ids = [name[len(OFFLINE_NAME_PREFIX):] for name in runs_names
                        if is_offline_run_name(name)]
    sync_offline_runs(base_path, project_name, offline_runs_ids)


def sync_all_runs(base_path: Path, project_name: Optional[str]) -> None:
    sync_all_registered_runs(base_path)

    offline_runs_ids = get_offline_runs_ids(base_path)
    sync_offline_runs(base_path, project_name, offline_runs_ids)


#######################################################################################################################
# Entrypoint for the CLI utility
#######################################################################################################################


# pylint: disable=unused-argument
def get_neptune_path(ctx, param, path: str) -> Path:
    # check if path exists and contains a '.neptune' folder
    path = Path(path)
    if (path / NEPTUNE_RUNS_DIRECTORY).is_dir():
        return path / NEPTUNE_RUNS_DIRECTORY
    elif path.name == NEPTUNE_RUNS_DIRECTORY and path.is_dir():
        return path
    else:
        raise click.BadParameter("Path {} does not contain a '{}' folder.".format(path, NEPTUNE_RUNS_DIRECTORY))


path_option = click.option('--path', type=click.Path(exists=True, file_okay=False, resolve_path=True),
                           default=Path.cwd(), callback=get_neptune_path, metavar='<location>',
                           help="path to a directory containing a '.neptune' folder with stored runs")


@click.command()
@path_option
def status(path: Path) -> None:
    """List synchronized and unsynchronized runs in the given directory. Trashed runs are not listed.

    Neptune stores run data on disk in '.neptune' directories. In case a run executes offline
    or network is unavailable as the run executes, run data can be synchronized
    with the server with this command line utility.

    Examples:

    \b
    # List synchronized and unsynchronized runs in the current directory
    neptune status

    \b
    # List synchronized and unsynchronized runs in directory "foo/bar" without actually syncing
    neptune status --path foo/bar
    """

    # pylint: disable=global-statement
    global backend
    backend = HostedNeptuneBackend(Credentials.from_token())

    synchronization_status(path)


@click.command()
@path_option
@click.option('--run', 'runs_names', multiple=True, metavar='<run-name>',
              help="run name (workspace/project/short-id or UUID for offline runs) to synchronize.")
@click.option('-p', '--project', 'project_name', multiple=False, metavar='project-name',
              help="project name (workspace/project) where offline runs will be sent")
def sync(path: Path, runs_names: List[str], project_name: Optional[str]):
    """Synchronizes runs with unsent data with the server.

    Neptune stores run data on disk in '.neptune' directories. In case a run executes offline
    or network is unavailable as the run executes, run data can be synchronized
    with the server with this command line utility.

    You can list unsynchronized runs with `neptune status`

    Examples:

    \b
    # Synchronize all runs in the current directory
    neptune sync

    \b
    # Synchronize all runs in the given path
    neptune sync --path foo/bar

    \b
    # Synchronize only runs "NPT-42" and "NPT-43" in "workspace/project" in the current directory
    neptune sync --run workspace/project/NPT-42 --run workspace/project/NPT-43

    \b
    # Synchronise all runs in the current directory, sending offline runs to project "workspace/project"
    neptune sync --project workspace/project

    \b
    # Synchronize only the offline run with UUID offline/a1561719-b425-4000-a65a-b5efb044d6bb
    # to project "workspace/project"
    neptune sync --project workspace/project --run offline/a1561719-b425-4000-a65a-b5efb044d6bb
    """

    # pylint: disable=global-statement
    global backend
    backend = HostedNeptuneBackend(Credentials.from_token())

    if runs_names:
        sync_selected_runs(path, project_name, runs_names)
    else:
        sync_all_runs(path, project_name)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import socket
import unittest
import uuid
from unittest.mock import call

from bravado.exception import HTTPNotFound
from mock import MagicMock, patch
from packaging.version import Version

from neptune.new.exceptions import CannotResolveHostname, FileUploadError, MetadataInconsistency, \
    UnsupportedClientVersion
from neptune.new.internal.backends.hosted_neptune_backend import HostedNeptuneBackend
from neptune.new.internal.backends.hosted_client import (
    DEFAULT_REQUEST_KWARGS,
    _get_token_client,  # pylint:disable=protected-access
    get_client_config,
    create_http_client_with_auth,
    create_backend_client,
    create_leaderboard_client,
    create_artifacts_client,
)
from neptune.new.internal.credentials import Credentials
from neptune.new.internal.operation import (
    AssignString,
    LogFloats,
    TrackFilesToArtifact,
    UploadFile,
    UploadFileContent
)
from neptune.new.internal.backends.utils import verify_host_resolution
from neptune.new.internal.utils import base64_encode
from tests.neptune.new.backend_test_mixin import BackendTestMixin

API_TOKEN = 'eyJhcGlfYWRkcmVzcyI6Imh0dHBzOi8vYXBwLnN0YWdlLm5lcHR1bmUubWwiLCJ' \
            'hcGlfa2V5IjoiOTJhNzhiOWQtZTc3Ni00ODlhLWI5YzEtNzRkYmI1ZGVkMzAyIn0='

credentials = Credentials.from_token(API_TOKEN)


@patch('neptune.new.internal.backends.hosted_client.RequestsClient', new=MagicMock())
@patch('neptune.new.internal.backends.hosted_client.NeptuneAuthenticator', new=MagicMock())
@patch('bravado.client.SwaggerClient.from_url')
@patch('platform.platform', new=lambda: 'testPlatform')
@patch('platform.python_version', new=lambda: '3.9.test')
class TestHostedNeptuneBackend(unittest.TestCase, BackendTestMixin):
    # pylint:disable=protected-access

    def setUp(self) -> None:
        # Clear all LRU storage
        verify_host_resolution.cache_clear()
        _get_token_client.cache_clear()
        get_client_config.cache_clear()
        create_http_client_with_auth.cache_clear()
        create_backend_client.cache_clear()
        create_leaderboard_client.cache_clear()
        create_artifacts_client.cache_clear()

    @patch('neptune.new.internal.backends.hosted_neptune_backend.upload_file_attribute')
    def test_execute_operations(self, upload_mock, swagger_client_factory):
        # given
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)
        backend = HostedNeptuneBackend(credentials)
        exp_uuid = str(uuid.uuid4())

        response_error = MagicMock()
        response_error.errorDescription = "error1"
        swagger_client.api.executeOperations().response().result = [response_error]
        swagger_client.api.executeOperations.reset_mock()
        upload_mock.return_value = FileUploadError("file1", "error2")
        some_text = "Some streamed text"
        some_binary = b"Some streamed binary"

        # when
        result = backend.execute_operations(
            run_id=exp_uuid,
            operations=[
                UploadFile(
                    path=['some', 'files', 'some_file'],
                    ext='',
                    file_path='path_to_file'
                ),
                UploadFileContent(
                    path=['some', 'files', 'some_text_stream'],
                    ext="txt",
                    file_content=base64_encode(some_text.encode('utf-8'))
                ),
                UploadFileContent(
                    path=['some', 'files', 'some_binary_stream'],
                    ext="bin",
                    file_content=base64_encode(some_binary)
                ),
                LogFloats(["images", "img1"], [LogFloats.ValueType(1, 2, 3)]),
                AssignString(["properties", "name"], "some text"),
                UploadFile(
                    path=['some', 'other', 'file.txt'],
                    ext="txt",
                    file_path='other/file/path.txt'
                )
            ]
        )

        # than
        swagger_client.api.executeOperations.assert_called_once_with(
            **{
                'experimentId': str(exp_uuid),
                'operations': [{
                    'path': "images/img1",
                    'logFloats': {
                        'entries': [{
                            'value': 1,
                            'step': 2,
                            'timestampMilliseconds': 3000
                        }]
                    }
                }, {
                    'path': "properties/name",
                    'assignString': {
                        'value': "some text"
                    }
                }],
                **DEFAULT_REQUEST_KWARGS,
            }
        )

        upload_mock.assert_has_calls([
            call(swagger_client=backend.leaderboard_client,
                 run_id=exp_uuid,
                 attribute="some/other/file.txt",
                 source="other/file/path.txt",
                 ext="txt"),
            call(swagger_client=backend.leaderboard_client,
                 run_id=exp_uuid,
                 attribute="some/files/some_file",
                 source="path_to_file",
                 ext=""),
            call(swagger_client=backend.leaderboard_client,
                 run_id=exp_uuid,
                 attribute="some/files/some_text_stream",
                 source=some_text.encode('utf-8'),
                 ext="txt"),
            call(swagger_client=backend.leaderboard_client,
                 run_id=exp_uuid,
                 attribute="some/files/some_binary_stream",
                 source=some_binary,
                 ext="bin")
        ], any_order=True)

        self.assertEqual([
            FileUploadError("file1", "error2"),
            FileUploadError("file1", "error2"),
            FileUploadError("file1", "error2"),
            FileUploadError("file1", "error2"),
            MetadataInconsistency("error1")
        ], result)

    @patch('neptune.new.internal.backends.hosted_neptune_backend.upload_file_attribute')
    def test_upload_files_destination_path(self, upload_mock, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory)
        backend = HostedNeptuneBackend(credentials)
        exp_uuid = str(uuid.uuid4())

        # when
        backend.execute_operations(
            run_id=exp_uuid,
            operations=[
                UploadFile(
                    path=['some', 'path', '1', "var"],
                    ext="",
                    file_path='/path/to/file'
                ),
                UploadFile(
                    path=['some', 'path', '2', "var"],
                    ext="txt",
                    file_path='/some.file/with.dots.txt'
                ),
                UploadFile(
                    path=['some', 'path', '3', "var"],
                    ext="jpeg",
                    file_path='/path/to/some_image.jpeg'
                )
            ]
        )

        upload_mock.assert_has_calls([
            call(swagger_client=backend.leaderboard_client,
                 run_id=exp_uuid,
                 attribute="some/path/1/var",
                 source="/path/to/file",
                 ext=""),
            call(swagger_client=backend.leaderboard_client,
                 run_id=exp_uuid,
                 attribute="some/path/2/var",
                 source="/some.file/with.dots.txt",
                 ext="txt"),
            call(swagger_client=backend.leaderboard_client,
                 run_id=exp_uuid,
                 attribute="some/path/3/var",
                 source="/path/to/some_image.jpeg",
                 ext="jpeg")
        ], any_order=True)

    @patch('neptune.new.internal.backends.hosted_neptune_backend.track_to_new_artifact')
    def test_track_to_new_artifact(self, track_to_new_artifact_mock, swagger_client_factory):
        # given
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)
        backend = HostedNeptuneBackend(credentials)
        exp_id = str(uuid.uuid4())
        project_id = str(uuid.uuid4())

        response_error = MagicMock()
        response_error.errorDescription = "error1"
        swagger_client.api.executeOperations.return_value.response.return_value.result = [response_error]
        swagger_client.api.getArtifactAttribute.side_effect = HTTPNotFound(response=MagicMock())

        # when
        backend.execute_operations(
            run_id=exp_id,
            operations=[
                TrackFilesToArtifact(
                    path=['sub', 'one'],
                    project_id=project_id,
                    entries=[('/path/to/file', '/path/to')]
                ),
                TrackFilesToArtifact(
                    path=['sub', 'two'],
                    project_id=project_id,
                    entries=[('/path/to/file1', None), ('/path/to/file2', None)]
                ),
                TrackFilesToArtifact(
                    path=['sub', 'three'],
                    project_id=project_id,
                    entries=[('/path/to/file1', None)]
                ),
                TrackFilesToArtifact(
                    path=['sub', 'three'],
                    project_id=project_id,
                    entries=[('/path/to/file2', None)]
                )
            ]
        )

        # then
        track_to_new_artifact_mock.assert_has_calls([
            call(swagger_client=swagger_client,
                 project_id=project_id,
                 path=["sub", "one"],
                 parent_identifier=str(exp_id),
                 entries=[("/path/to/file", '/path/to')],
                 default_request_params=DEFAULT_REQUEST_KWARGS),
            call(swagger_client=swagger_client,
                 project_id=project_id,
                 path=["sub", "two"],
                 parent_identifier=str(exp_id),
                 entries=[("/path/to/file1", None), ("/path/to/file2", None)],
                 default_request_params=DEFAULT_REQUEST_KWARGS),
            call(swagger_client=swagger_client,
                 project_id=project_id,
                 path=["sub", "three"],
                 parent_identifier=str(exp_id),
                 entries=[("/path/to/file1", None), ("/path/to/file2", None)],
                 default_request_params=DEFAULT_REQUEST_KWARGS),
        ], any_order=True)

    @patch('neptune.new.internal.backends.hosted_neptune_backend.track_to_existing_artifact')
    def test_track_to_existing_artifact(self, track_to_existing_artifact_mock, swagger_client_factory):
        # given
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)
        backend = HostedNeptuneBackend(credentials)
        exp_id = str(uuid.uuid4())
        project_id = str(uuid.uuid4())

        response_error = MagicMock()
        response_error.errorDescription = "error1"
        swagger_client.api.executeOperations.return_value.response.return_value.result = [response_error]
        swagger_client.api.getArtifactAttribute.return_value.response.return_value.result.hash = 'dummyHash'

        # when
        backend.execute_operations(
            run_id=exp_id,
            operations=[
                TrackFilesToArtifact(
                    path=['sub', 'one'],
                    project_id=project_id,
                    entries=[('/path/to/file', '/path/to')]
                ),
                TrackFilesToArtifact(
                    path=['sub', 'two'],
                    project_id=project_id,
                    entries=[('/path/to/file1', None), ('/path/to/file2', None)]
                ),
                TrackFilesToArtifact(
                    path=['sub', 'three'],
                    project_id=project_id,
                    entries=[('/path/to/file1', None)]
                ),
                TrackFilesToArtifact(
                    path=['sub', 'three'],
                    project_id=project_id,
                    entries=[('/path/to/file2', None)]
                )
            ]
        )

        # then
        track_to_existing_artifact_mock.assert_has_calls([
            call(swagger_client=swagger_client,
                 project_id=project_id,
                 path=["sub", "one"],
                 artifact_hash='dummyHash',
                 parent_identifier=str(exp_id),
                 entries=[("/path/to/file", '/path/to')],
                 default_request_params=DEFAULT_REQUEST_KWARGS),
            call(swagger_client=swagger_client,
                 project_id=project_id,
                 path=["sub", "two"],
                 artifact_hash='dummyHash',
                 parent_identifier=str(exp_id),
                 entries=[("/path/to/file1", None), ("/path/to/file2", None)],
                 default_request_params=DEFAULT_REQUEST_KWARGS),
            call(swagger_client=swagger_client,
                 project_id=project_id,
                 path=["sub", "three"],
                 artifact_hash='dummyHash',
                 parent_identifier=str(exp_id),
                 entries=[("/path/to/file1", None), ("/path/to/file2", None)],
                 default_request_params=DEFAULT_REQUEST_KWARGS),
        ], any_order=True)

    @patch('neptune.new.internal.backends.hosted_client.neptune_client_version', Version('0.5.13'))
    def test_min_compatible_version_ok(self, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory, min_compatible='0.5.13')

        # expect
        HostedNeptuneBackend(credentials)

    @patch('neptune.new.internal.backends.hosted_client.neptune_client_version', Version('0.5.13'))
    def test_min_compatible_version_fail(self, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory, min_compatible='0.5.14')

        # expect
        with self.assertRaises(UnsupportedClientVersion) as ex:
            HostedNeptuneBackend(credentials)

        self.assertTrue("Please install neptune-client>=0.5.14" in str(ex.exception))

    @patch('neptune.new.internal.backends.hosted_client.neptune_client_version', Version('0.5.13'))
    def test_max_compatible_version_ok(self, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory, max_compatible='0.5.12')

        # expect
        HostedNeptuneBackend(credentials)

    @patch('neptune.new.internal.backends.hosted_client.neptune_client_version', Version('0.5.13'))
    def test_max_compatible_version_fail(self, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory, max_compatible='0.4.999')

        # expect
        with self.assertRaises(UnsupportedClientVersion) as ex:
            HostedNeptuneBackend(credentials)

        self.assertTrue("Please install neptune-client==0.4.0" in str(ex.exception))

    @patch('socket.gethostbyname')
    def test_cannot_resolve_host(self, gethostname_mock, _):
        # given
        gethostname_mock.side_effect = socket.gaierror

        # expect
        with self.assertRaises(CannotResolveHostname):
            HostedNeptuneBackend(credentials)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from neptune.utils import NoopObject


class TestHostedNeptuneObject(unittest.TestCase):

    def test_builtin_fields_not_overriden(self):
        # given
        objectUnderTest = NoopObject()

        # then
        self.assertIsInstance(objectUnderTest.__class__, type)

    def test_attributes_fall_back_on_getattr(self):
        # given
        objectUnderTest = NoopObject()
        some_value = 42

        # then
        self.assertEqual(objectUnderTest.foo, objectUnderTest)
        self.assertEqual(objectUnderTest.bar(some_value), objectUnderTest)

    def test_attributes_fall_back_on_getitem(self):
        # given
        objectUnderTest = NoopObject()
        some_value = 42

        # then
        self.assertEqual(objectUnderTest['foo'], objectUnderTest)
        self.assertEqual(objectUnderTest['bar'](some_value), objectUnderTest)

    def test_noop_object_callable(self):
        # given
        objectUnderTest = NoopObject()
        some_value = 42

        # then
        self.assertEqual(objectUnderTest(some_value), objectUnderTest)

    def test_noop_object_context_manager(self):
        # given
        objectUnderTest = NoopObject()

        # when
        with objectUnderTest as e:

            # then
            e(42)



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

REGISTERED_CODES = dict()


class ManagementOperationFailure(Exception):
    code = -1
    description = "Unknown error"

    def __init__(self, **kwargs):
        super().__init__()
        self._properties: dict = kwargs or {}

    def __str__(self):
        return f"{self.description.format(**self._properties)} (code: {self.code})"

    def __init_subclass__(cls):
        previous = REGISTERED_CODES.get(cls.code)
        assert previous is None, f"{cls} cannot have code {cls.code} already used by {previous}"
        REGISTERED_CODES[cls.code] = cls

    @property
    def details(self):
        return {
            "code": self.code,
            "description": self.description.format(**self._properties),
        }


class InvalidProjectName(ManagementOperationFailure):
    code = 1
    description = 'Provided project name "{name}" could not be parsed.'


class MissingWorkspaceName(ManagementOperationFailure):
    code = 2
    description = 'Cannot resolve project "{name}", you have to provide a workspace name.'


class ConflictingWorkspaceName(ManagementOperationFailure):
    code = 3
    description = 'Project name "{name}" conflicts with provided workspace "{workspace}".'


class ProjectNotFound(ManagementOperationFailure):
    code = 4
    description = 'Project "{name}" could not be found.'


class WorkspaceNotFound(ManagementOperationFailure):
    code = 5
    description = 'Workspace "{workspace}" could not be found.'


class ProjectAlreadyExists(ManagementOperationFailure):
    code = 6
    description = 'Project "{name}" already exists.'


class AccessRevokedOnDeletion(ManagementOperationFailure):
    code = 7
    description = 'Not enough permissions to delete project "{name}".'


class AccessRevokedOnMemberRemoval(ManagementOperationFailure):
    code = 8
    description = 'Not enough permissions to remove user "{user}" from project "{project}".'


class UserNotExistsOrWithoutAccess(ManagementOperationFailure):
    code = 9
    description = 'User "{user}" not exists or has no access to project "{project}".'


class UserAlreadyHasAccess(ManagementOperationFailure):
    code = 10
    description = 'User "{user}" already has access to project "{project}".'


class ProjectsLimitReached(ManagementOperationFailure):
    code = 11
    description = 'Project number limit reached.'


class BadRequestException(ManagementOperationFailure):
    code = 400
    description = 'Your request has encountered following validation errors: {validation_errors}'



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import abc
import enum
import typing
import pathlib
from dataclasses import dataclass

from neptune.new.exceptions import NeptuneUnhandledArtifactSchemeException, NeptuneUnhandledArtifactTypeException


class ArtifactFileType(enum.Enum):
    S3 = "S3"
    LOCAL = "Local"


class ArtifactMetadataSerializer:
    @staticmethod
    def serialize(metadata: typing.Dict[str, str]) -> typing.List[typing.Dict[str, str]]:
        return [{"key": k, "value": v} for k, v in sorted(metadata.items())]

    @staticmethod
    def deserialize(metadata: typing.List[typing.Dict[str, str]]) -> typing.Dict[str, str]:
        return {
            f'{key_value.get("key")}': f'{key_value.get("value")}'
            for key_value in metadata
        }


@dataclass
class ArtifactFileData:
    file_path: str
    file_hash: str
    type: str
    metadata: typing.Dict[str, str]
    size: int = None

    @classmethod
    def from_dto(cls, artifact_file_dto):
        return cls(
            file_path=artifact_file_dto.filePath,
            file_hash=artifact_file_dto.fileHash,
            type=artifact_file_dto.type,
            size=artifact_file_dto.size,
            metadata=ArtifactMetadataSerializer.deserialize(
                [
                    {'key': str(m.key), 'value': str(m.value)}
                    for m in artifact_file_dto.metadata
                ]
            )
        )

    def to_dto(self) -> typing.Dict:
        return {
            'filePath': self.file_path,
            'fileHash': self.file_hash,
            'type': self.type,
            'size': self.size,
            'metadata': ArtifactMetadataSerializer.serialize(self.metadata)
        }


class ArtifactDriversMap:
    _implementations: typing.List[typing.Type['ArtifactDriver']] = []

    @classmethod
    def match_path(cls, path: str) -> typing.Type['ArtifactDriver']:
        for artifact_driver in cls._implementations:
            if artifact_driver.matches(path):
                return artifact_driver

        raise NeptuneUnhandledArtifactSchemeException(path)

    @classmethod
    def match_type(cls, type_str: str) -> typing.Type['ArtifactDriver']:
        for artifact_driver in cls._implementations:
            if artifact_driver.get_type() == type_str:
                return artifact_driver

        raise NeptuneUnhandledArtifactTypeException(type_str)


class ArtifactDriver(abc.ABC):
    def __init_subclass__(cls):
        # pylint: disable=protected-access
        ArtifactDriversMap._implementations.append(cls)

    @staticmethod
    def get_type() -> str:
        raise NotImplementedError

    @classmethod
    def matches(cls, path: str) -> bool:
        raise NotImplementedError

    @classmethod
    def get_tracked_files(cls, path: str, destination: str = None) -> typing.List[ArtifactFileData]:
        raise NotImplementedError

    @classmethod
    def download_file(cls, destination: pathlib.Path, file_definition: ArtifactFileData):
        raise NotImplementedError



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import typing
import pathlib
from datetime import datetime
from urllib.parse import urlparse

import boto3
from botocore.exceptions import NoCredentialsError

from neptune.new.internal.artifacts.types import ArtifactDriver, ArtifactFileData, ArtifactFileType
from neptune.new.exceptions import (
    NeptuneRemoteStorageAccessException,
    NeptuneRemoteStorageCredentialsException,
    NeptuneUnsupportedArtifactFunctionalityException,
)


class S3ArtifactDriver(ArtifactDriver):
    DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"

    @staticmethod
    def get_type() -> str:
        return ArtifactFileType.S3.value

    @classmethod
    def matches(cls, path: str) -> bool:
        return urlparse(path).scheme == 's3'

    @classmethod
    def _serialize_metadata(cls, metadata: typing.Dict[str, typing.Any]) -> typing.Dict[str, str]:
        return {
            "location": metadata['location'],
            "last_modified": metadata['last_modified'].strftime(cls.DATETIME_FORMAT),
        }

    @classmethod
    def _deserialize_metadata(cls, metadata: typing.Dict[str, str]) -> typing.Dict[str, typing.Any]:
        return {
            "location": metadata['location'],
            "last_modified": datetime.strptime(metadata['last_modified'], cls.DATETIME_FORMAT),
        }

    @classmethod
    def get_tracked_files(cls, path: str, destination: str = None) -> typing.List[ArtifactFileData]:
        url = urlparse(path)
        bucket_name, prefix = url.netloc, url.path.lstrip('/')

        if '*' in prefix:
            raise NeptuneUnsupportedArtifactFunctionalityException(
                f'Wildcard characters (*,?) in location URI ({path}) are not supported.'
            )

        # pylint: disable=no-member
        remote_storage = boto3.resource('s3').Bucket(bucket_name)

        stored_files: typing.List[ArtifactFileData] = list()

        try:
            for remote_object in remote_storage.objects.filter(Prefix=prefix):
                # If prefix is path to file get only directories
                if prefix == remote_object.key:
                    prefix = str(pathlib.PurePosixPath(prefix).parent)

                remote_key = remote_object.key
                destination = pathlib.PurePosixPath(destination or '')
                relative_file_path = remote_key[len(prefix.lstrip('.')):].lstrip("/")

                file_path = destination / relative_file_path

                stored_files.append(
                    ArtifactFileData(
                        file_path=str(file_path).lstrip('/'),
                        file_hash=remote_object.e_tag.strip('"'),
                        type=ArtifactFileType.S3.value,
                        size=remote_object.size,
                        metadata=cls._serialize_metadata({
                            "location": f's3://{bucket_name}/{remote_key.lstrip("/")}',
                            "last_modified": remote_object.last_modified,
                        })
                    )
                )
        except NoCredentialsError:
            raise NeptuneRemoteStorageCredentialsException()
        except (remote_storage.meta.client.exceptions.NoSuchBucket,
                remote_storage.meta.client.exceptions.NoSuchKey):
            raise NeptuneRemoteStorageAccessException(location=path)

        return stored_files

    @classmethod
    def download_file(cls, destination: pathlib.Path, file_definition: ArtifactFileData):
        location = file_definition.metadata.get('location')
        url = urlparse(location)
        bucket_name, path = url.netloc, url.path.lstrip('/')

        remote_storage = boto3.resource('s3')
        try:
            # pylint: disable=no-member
            bucket = remote_storage.Bucket(bucket_name)
            bucket.download_file(path, str(destination))
        except NoCredentialsError:
            raise NeptuneRemoteStorageCredentialsException()
        except (remote_storage.meta.client.exceptions.NoSuchBucket,
                remote_storage.meta.client.exceptions.NoSuchKey):
            raise NeptuneRemoteStorageAccessException(location=location)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import uuid
from datetime import datetime
from shutil import copyfile
from typing import Optional, List, Dict, TypeVar, Type, Tuple, Any
from zipfile import ZipFile

from neptune.new.exceptions import (
    RunNotFound,
    RunUUIDNotFound,
    InternalClientError,
    MetadataInconsistency,
    NeptuneException,
)
from neptune.new.internal.artifacts.types import ArtifactFileData
from neptune.new.internal.backends.api_model import (
    ApiRun,
    ArtifactAttribute,
    Attribute,
    AttributeType,
    BoolAttribute,
    DatetimeAttribute,
    FileAttribute,
    FloatAttribute,
    FloatSeriesAttribute,
    IntAttribute,
    Project,
    Workspace,
    StringAttribute,
    StringSeriesAttribute,
    StringSetAttribute,
    StringSeriesValues,
    FloatSeriesValues,
    ImageSeriesValues,
    StringPointValue,
)
from neptune.new.internal.backends.hosted_file_operations import get_unique_upload_entries
from neptune.new.internal.backends.neptune_backend import NeptuneBackend
from neptune.new.internal.run_structure import RunStructure
from neptune.new.internal.operation import (
    AssignArtifact,
    AddStrings,
    AssignBool,
    AssignDatetime,
    AssignFloat,
    AssignInt,
    AssignString,
    ClearArtifact,
    ClearFloatLog,
    ClearImageLog,
    ClearStringLog,
    ClearStringSet,
    ConfigFloatSeries,
    DeleteAttribute,
    DeleteFiles,
    LogFloats,
    LogImages,
    LogStrings,
    Operation,
    RemoveStrings,
    TrackFilesToArtifact,
    UploadFile,
    UploadFileContent,
    UploadFileSet,

)
from neptune.new.internal.operation_visitor import OperationVisitor
from neptune.new.internal.utils import base64_decode
from neptune.new.internal.utils.generic_attribute_mapper import NoValue
from neptune.new.internal.utils.paths import path_to_str
from neptune.new.types import Boolean, Integer
from neptune.new.types.atoms import GitRef
from neptune.new.types.atoms.artifact import Artifact
from neptune.new.types.atoms.datetime import Datetime
from neptune.new.types.atoms.file import File
from neptune.new.types.atoms.float import Float
from neptune.new.types.atoms.string import String
from neptune.new.types.file_set import FileSet
from neptune.new.types.namespace import Namespace
from neptune.new.types.series.float_series import FloatSeries
from neptune.new.types.series.file_series import FileSeries
from neptune.new.types.series.string_series import StringSeries
from neptune.new.types.sets.string_set import StringSet
from neptune.new.types.value import Value
from neptune.new.types.value_visitor import ValueVisitor

Val = TypeVar('Val', bound=Value)


class NeptuneBackendMock(NeptuneBackend):

    def close(self) -> None:
        pass

    def __init__(self, credentials=None, proxies=None):
        # pylint: disable=unused-argument
        self._runs: Dict[str, RunStructure[Value, dict]] = dict()
        self._artifacts: Dict[Tuple[str, str], List[ArtifactFileData]] = dict()
        self._attribute_type_converter_value_visitor = self.AttributeTypeConverterValueVisitor()

    def get_display_address(self) -> str:
        return "OFFLINE"

    def get_project(self, project_id: str) -> Project:
        return Project(str(uuid.uuid4()), "project-placeholder", "offline")

    def get_available_projects(self,
                               workspace_id: Optional[str] = None,
                               search_term: Optional[str] = None
                               ) -> List[Project]:
        return [Project(str(uuid.uuid4()), "project-placeholder", "offline")]

    def get_available_workspaces(self) -> List[Workspace]:
        return [Workspace(str(uuid.uuid4()), "offline")]

    def create_run(self,
                   project_id: str,
                   git_ref: Optional[GitRef] = None,
                   custom_run_id: Optional[str] = None,
                   notebook_id: Optional[str] = None,
                   checkpoint_id: Optional[str] = None
                   ) -> ApiRun:
        short_id = "OFFLINE-{}".format(len(self._runs) + 1)
        new_run_id = str(uuid.uuid4())
        self._runs[new_run_id] = RunStructure[Value, dict]()
        self._runs[new_run_id].set(["sys", "id"], String(short_id))
        self._runs[new_run_id].set(["sys", "state"], String("running"))
        self._runs[new_run_id].set(["sys", "owner"], String("offline_user"))
        self._runs[new_run_id].set(["sys", "size"], Float(0))
        self._runs[new_run_id].set(["sys", "tags"], StringSet(set()))
        self._runs[new_run_id].set(["sys", "creation_time"], Datetime(datetime.now()))
        self._runs[new_run_id].set(["sys", "modification_time"], Datetime(datetime.now()))
        self._runs[new_run_id].set(["sys", "failed"], Boolean(False))
        if git_ref:
            self._runs[new_run_id].set(["source_code", "git"], git_ref)
        return ApiRun(new_run_id, short_id, 'workspace', 'sandbox', False)

    def create_checkpoint(self, notebook_id: str, jupyter_path: str) -> Optional[str]:
        return None

    def get_run(self, run_id: str) -> ApiRun:
        raise RunNotFound(run_id)

    def execute_operations(self, run_id: str, operations: List[Operation]) -> List[NeptuneException]:
        result = []
        for op in operations:
            try:
                self._execute_operation(run_id, op)
            except NeptuneException as e:
                result.append(e)
        return result

    def _execute_operation(self, run_id: str, op: Operation) -> None:
        if run_id not in self._runs:
            raise RunUUIDNotFound(run_id)
        run = self._runs[run_id]
        val = run.get(op.path)
        if val is not None and not isinstance(val, Value):
            if isinstance(val, dict):
                raise MetadataInconsistency("{} is a namespace, not an attribute".format(op.path))
            else:
                raise InternalClientError("{} is a {}".format(op.path, type(val)))
        visitor = NeptuneBackendMock.NewValueOpVisitor(op.path, val)
        new_val = visitor.visit(op)
        if new_val is not None:
            run.set(op.path, new_val)
        else:
            run.pop(op.path)

    def get_attributes(self, run_id: str) -> List[Attribute]:
        if run_id not in self._runs:
            raise RunUUIDNotFound(run_id)
        run = self._runs[run_id]
        return list(self._generate_attributes(None, run.get_structure()))

    def _generate_attributes(self, base_path: Optional[str], values: dict):
        for key, value_or_dict in values.items():
            new_path = base_path + '/' + key if base_path is not None else key
            if isinstance(value_or_dict, dict):
                yield from self._generate_attributes(new_path, value_or_dict)
            else:
                yield Attribute(new_path, value_or_dict.accept(self._attribute_type_converter_value_visitor))

    def download_file(self, run_id: str, path: List[str], destination: Optional[str] = None):
        value: File = self._runs[run_id].get(path)
        target_path = os.path.abspath(destination or (path[-1] + ("." + value.extension if value.extension else "")))
        if value.content is not None:
            with open(target_path, 'wb') as target_file:
                target_file.write(value.content)
        elif value.path != target_path:
            copyfile(value.path, target_path)

    def download_file_set(self, run_id: str, path: List[str], destination: Optional[str] = None):
        source_file_set_value: FileSet = self._runs[run_id].get(path)

        if destination is None:
            target_file = path[-1] + ".zip"
        elif os.path.isdir(destination):
            target_file = os.path.join(destination, path[-1] + ".zip")
        else:
            target_file = destination

        upload_entries = get_unique_upload_entries(source_file_set_value.file_globs)

        with ZipFile(target_file, 'w') as zipObj:
            for upload_entry in upload_entries:
                zipObj.write(upload_entry.source_path, upload_entry.target_path)

    def get_float_attribute(self, run_id: str, path: List[str]) -> FloatAttribute:
        val = self._get_attribute(run_id, path, Float)
        return FloatAttribute(val.value)

    def get_int_attribute(self, run_id: str, path: List[str]) -> IntAttribute:
        val = self._get_attribute(run_id, path, Integer)
        return IntAttribute(val.value)

    def get_bool_attribute(self, run_id: str, path: List[str]) -> BoolAttribute:
        val = self._get_attribute(run_id, path, Boolean)
        return BoolAttribute(val.value)

    def get_file_attribute(self, run_id: str, path: List[str]) -> FileAttribute:
        val = self._get_attribute(run_id, path, File)
        return FileAttribute(
            name=os.path.basename(val.path) if val.path else "",
            ext=val.extension or "",
            size=0)

    def get_string_attribute(self, run_id: str, path: List[str]) -> StringAttribute:
        val = self._get_attribute(run_id, path, String)
        return StringAttribute(val.value)

    def get_datetime_attribute(self, run_id: str, path: List[str]) -> DatetimeAttribute:
        val = self._get_attribute(run_id, path, Datetime)
        return DatetimeAttribute(val.value)

    def get_artifact_attribute(self, run_id: str, path: List[str]) -> ArtifactAttribute:
        val = self._get_attribute(run_id, path, Artifact)
        return ArtifactAttribute(val.hash)

    def list_artifact_files(self, project_id: str, artifact_hash: str) -> List[ArtifactFileData]:
        return self._artifacts[(project_id, artifact_hash)]

    def get_float_series_attribute(self, run_id: str, path: List[str]) -> FloatSeriesAttribute:
        val = self._get_attribute(run_id, path, FloatSeries)
        return FloatSeriesAttribute(val.values[-1] if val.values else None)

    def get_string_series_attribute(self, run_id: str, path: List[str]) -> StringSeriesAttribute:
        val = self._get_attribute(run_id, path, StringSeries)
        return StringSeriesAttribute(val.values[-1] if val.values else None)

    def get_string_set_attribute(self, run_id: str, path: List[str]) -> StringSetAttribute:
        val = self._get_attribute(run_id, path, StringSet)
        return StringSetAttribute(set(val.values))

    def _get_attribute(self, run_id: str, path: List[str], expected_type: Type[Val]) -> Val:
        if run_id not in self._runs:
            raise RunUUIDNotFound(run_id)
        value: Optional[Value] = self._runs[run_id].get(path)
        str_path = path_to_str(path)
        if value is None:
            raise MetadataInconsistency("Attribute {} not found".format(str_path))
        if isinstance(value, expected_type):
            return value
        raise MetadataInconsistency("Attribute {} is not {}".format(str_path, type.__name__))

    def get_string_series_values(self, run_id: str, path: List[str],
                                 offset: int, limit: int) -> StringSeriesValues:
        val = self._get_attribute(run_id, path, StringSeries)
        return StringSeriesValues(
            len(val.values),
            [StringPointValue(timestampMillis=42342, step=idx, value=v) for idx, v in enumerate(val.values)]
        )

    def get_float_series_values(self, run_id: str, path: List[str],
                                offset: int, limit: int) -> FloatSeriesValues:
        return FloatSeriesValues(0, [])

    def get_image_series_values(self, run_id: str, path: List[str],
                                offset: int, limit: int) -> ImageSeriesValues:
        return ImageSeriesValues(0)

    def download_file_series_by_index(self, run_id: str, path: List[str],
                                      index: int, destination: str):
        pass

    def get_run_url(self, run_id: str, workspace: str, project_name: str, short_id: str) -> str:
        return f"offline/{run_id}"

    def _get_attribute_values(self, value_dict, path_prefix: List[str]):
        assert isinstance(value_dict, dict)
        for k, value in value_dict.items():
            if isinstance(value, dict):
                yield from self._get_attribute_values(value, path_prefix + [k])
            else:
                attr_type = value.accept(self._attribute_type_converter_value_visitor).value
                attr_path = '/'.join(path_prefix + [k])
                if hasattr(value, "value"):
                    yield attr_path, attr_type, value.value
                else:
                    return attr_path, attr_type, NoValue

    def fetch_atom_attribute_values(self, run_id: str, path: List[str]) -> List[Tuple[str, AttributeType, Any]]:
        values = self._get_attribute_values(self._runs[run_id].get(path), path)
        namespace_prefix = path_to_str(path)
        if namespace_prefix:
            # don't want to catch "ns/attribute/other" while looking for "ns/attr"
            namespace_prefix += "/"
        return [
            (full_path, attr_type, attr_value)
            for (full_path, attr_type, attr_value) in values
            if full_path.startswith(namespace_prefix)
        ]

    class AttributeTypeConverterValueVisitor(ValueVisitor[AttributeType]):

        def visit_float(self, _: Float) -> AttributeType:
            return AttributeType.FLOAT

        def visit_integer(self, _: Integer) -> AttributeType:
            return AttributeType.INT

        def visit_boolean(self, _: Boolean) -> AttributeType:
            return AttributeType.BOOL

        def visit_string(self, _: String) -> AttributeType:
            return AttributeType.STRING

        def visit_datetime(self, _: Datetime) -> AttributeType:
            return AttributeType.DATETIME

        def visit_file(self, _: File) -> AttributeType:
            return AttributeType.FILE

        def visit_file_set(self, _: FileSet) -> AttributeType:
            return AttributeType.FILE_SET

        def visit_float_series(self, _: FloatSeries) -> AttributeType:
            return AttributeType.FLOAT_SERIES

        def visit_string_series(self, _: StringSeries) -> AttributeType:
            return AttributeType.STRING_SERIES

        def visit_image_series(self, _: FileSeries) -> AttributeType:
            return AttributeType.IMAGE_SERIES

        def visit_string_set(self, _: StringSet) -> AttributeType:
            return AttributeType.STRING_SET

        def visit_git_ref(self, _: GitRef) -> AttributeType:
            return AttributeType.GIT_REF

        def visit_artifact(self, _: Artifact) -> AttributeType:
            return AttributeType.ARTIFACT

        def visit_namespace(self, _: Namespace) -> AttributeType:
            raise NotImplementedError

    class NewValueOpVisitor(OperationVisitor[Optional[Value]]):

        def __init__(self, path: List[str], current_value: Optional[Value]):
            self._path = path
            self._current_value = current_value
            self._artifact_hash = 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'

        def visit_assign_float(self, op: AssignFloat) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Float):
                raise self._create_type_error("assign", Float.__name__)
            return Float(op.value)

        def visit_assign_int(self, op: AssignInt) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Integer):
                raise self._create_type_error("assign", Integer.__name__)
            return Integer(op.value)

        def visit_assign_bool(self, op: AssignBool) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Boolean):
                raise self._create_type_error("assign", Boolean.__name__)
            return Boolean(op.value)

        def visit_assign_string(self, op: AssignString) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, String):
                raise self._create_type_error("assign", String.__name__)
            return String(op.value)

        def visit_assign_datetime(self, op: AssignDatetime) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Datetime):
                raise self._create_type_error("assign", Datetime.__name__)
            return Datetime(op.value)

        def visit_assign_artifact(self, op: AssignArtifact) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Artifact):
                raise self._create_type_error("assign", Artifact.__name__)
            return Artifact(op.hash)

        def visit_track_files_to_artifact(self, _: TrackFilesToArtifact) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Artifact):
                raise self._create_type_error("save", Artifact.__name__)
            return Artifact(self._artifact_hash)

        def visit_clear_artifact(self, _: ClearArtifact) -> Optional[Value]:
            if self._current_value is None:
                return Artifact()
            if not isinstance(self._current_value, Artifact):
                raise self._create_type_error("clear", Artifact.__name__)
            return Artifact()

        def visit_upload_file(self, op: UploadFile) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, File):
                raise self._create_type_error("save", File.__name__)
            return File(path=op.file_path, extension=op.ext)

        def visit_upload_file_content(self, op: UploadFileContent) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, File):
                raise self._create_type_error("upload_files", File.__name__)
            return File(content=base64_decode(op.file_content), extension=op.ext)

        def visit_upload_file_set(self, op: UploadFileSet) -> Optional[Value]:
            if self._current_value is None or op.reset:
                return FileSet(op.file_globs)
            if not isinstance(self._current_value, FileSet):
                raise self._create_type_error("save", FileSet.__name__)
            return FileSet(self._current_value.file_globs + op.file_globs)

        def visit_log_floats(self, op: LogFloats) -> Optional[Value]:
            raw_values = [x.value for x in op.values]
            if self._current_value is None:
                return FloatSeries(raw_values)
            if not isinstance(self._current_value, FloatSeries):
                raise self._create_type_error("log", FloatSeries.__name__)
            return FloatSeries(self._current_value.values + raw_values,
                               min=self._current_value.min,
                               max=self._current_value.max,
                               unit=self._current_value.unit)

        def visit_log_strings(self, op: LogStrings) -> Optional[Value]:
            raw_values = [x.value for x in op.values]
            if self._current_value is None:
                return StringSeries(raw_values)
            if not isinstance(self._current_value, StringSeries):
                raise self._create_type_error("log", StringSeries.__name__)
            return StringSeries(self._current_value.values + raw_values)

        def visit_log_images(self, op: LogImages) -> Optional[Value]:
            raw_values = [File.from_content(base64_decode(x.value.data)) for x in op.values]
            if self._current_value is None:
                return FileSeries(raw_values)
            if not isinstance(self._current_value, FileSeries):
                raise self._create_type_error("log", FileSeries.__name__)
            return FileSeries(self._current_value.values + raw_values)

        def visit_clear_float_log(self, op: ClearFloatLog) -> Optional[Value]:
            # pylint: disable=unused-argument
            if self._current_value is None:
                return FloatSeries([])
            if not isinstance(self._current_value, FloatSeries):
                raise self._create_type_error("clear", FloatSeries.__name__)
            return FloatSeries([],
                               min=self._current_value.min,
                               max=self._current_value.max,
                               unit=self._current_value.unit)

        def visit_clear_string_log(self, op: ClearStringLog) -> Optional[Value]:
            # pylint: disable=unused-argument
            if self._current_value is None:
                return StringSeries([])
            if not isinstance(self._current_value, StringSeries):
                raise self._create_type_error("clear", StringSeries.__name__)
            return StringSeries([])

        def visit_clear_image_log(self, op: ClearImageLog) -> Optional[Value]:
            # pylint: disable=unused-argument
            if self._current_value is None:
                return FileSeries([])
            if not isinstance(self._current_value, FileSeries):
                raise self._create_type_error("clear", FileSeries.__name__)
            return FileSeries([])

        def visit_config_float_series(self, op: ConfigFloatSeries) -> Optional[Value]:
            if self._current_value is None:
                return FloatSeries([], min=op.min, max=op.max, unit=op.unit)
            if not isinstance(self._current_value, FloatSeries):
                raise self._create_type_error("log", FloatSeries.__name__)
            return FloatSeries(self._current_value.values, min=op.min, max=op.max, unit=op.unit)

        def visit_add_strings(self, op: AddStrings) -> Optional[Value]:
            if self._current_value is None:
                return StringSet(op.values)
            if not isinstance(self._current_value, StringSet):
                raise self._create_type_error("add", StringSet.__name__)
            return StringSet(self._current_value.values.union(op.values))

        def visit_remove_strings(self, op: RemoveStrings) -> Optional[Value]:
            if self._current_value is None:
                return StringSet(set())
            if not isinstance(self._current_value, StringSet):
                raise self._create_type_error("remove", StringSet.__name__)
            return StringSet(self._current_value.values.difference(op.values))

        def visit_clear_string_set(self, op: ClearStringSet) -> Optional[Value]:
            # pylint: disable=unused-argument
            if self._current_value is None:
                return StringSet(set())
            if not isinstance(self._current_value, StringSet):
                raise self._create_type_error("clear", StringSet.__name__)
            return StringSet(set())

        def visit_delete_files(self, op: DeleteFiles) -> Optional[Value]:
            # pylint: disable=unused-argument
            if self._current_value is None:
                return FileSet([])
            if not isinstance(self._current_value, FileSet):
                raise self._create_type_error("delete_files", FileSet.__name__)
            # It is not important to support deleting properly in debug mode, let's just ignore this operation
            return self._current_value

        def visit_delete_attribute(self, op: DeleteAttribute) -> Optional[Value]:
            # pylint: disable=unused-argument
            if self._current_value is None:
                raise MetadataInconsistency(
                    "Cannot perform delete operation on {}. Attribute is undefined.".format(self._path))
            return None

        def _create_type_error(self, op_name, expected):
            return MetadataInconsistency("Cannot perform {} operation on {}. Expected {}, {} found."
                                         .format(op_name, self._path, expected, type(self._current_value)))



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class ChannelWithLastValue:
    def __init__(self, channel_with_value_dto):
        self.channel_with_value_dto = channel_with_value_dto

    @property
    def id(self):
        return self.channel_with_value_dto.channelId

    @property
    def name(self):
        return self.channel_with_value_dto.channelName

    @property
    def type(self):
        return self.channel_with_value_dto.channelType

    @property
    def x(self):
        return self.channel_with_value_dto.x

    @x.setter
    def x(self, x):
        self.channel_with_value_dto.x = x

    @property
    def trimmed_y(self):
        return self.y[:255] if self.type == 'text' else self.y

    @property
    def y(self):
        return self.channel_with_value_dto.y

    @y.setter
    def y(self, y):
        self.channel_with_value_dto.y = y


class LeaderboardEntry(object):
    def __init__(self, project_leaderboard_entry_dto):
        self.project_leaderboard_entry_dto = project_leaderboard_entry_dto

    @property
    def id(self):
        return self.project_leaderboard_entry_dto.shortId

    @property
    def name(self):
        return self.project_leaderboard_entry_dto.name

    @property
    def state(self):
        return self.project_leaderboard_entry_dto.state

    @property
    def internal_id(self):
        return self.project_leaderboard_entry_dto.id

    @property
    def project_full_id(self):
        return '{org_name}/{project_name}'.format(
            org_name=self.project_leaderboard_entry_dto.organizationName,
            project_name=self.project_leaderboard_entry_dto.projectName)

    @property
    def system_properties(self):
        entry = self.project_leaderboard_entry_dto
        return {
            'id': entry.shortId,
            'name': entry.name,
            'created': entry.timeOfCreation,
            'finished': entry.timeOfCompletion,
            'running_time': entry.runningTime,
            'owner': entry.owner,
            'size': entry.size,
            'tags': entry.tags,
            'notes': entry.description
        }

    @property
    def channels(self):
        return [ChannelWithLastValue(ch) for ch in self.project_leaderboard_entry_dto.channelsLastValues]

    def add_channel(self, channel):
        self.project_leaderboard_entry_dto.channelsLastValues.append(channel.channel_with_value_dto)

    @property
    def channels_dict_by_name(self):
        return dict(
            (ch.name, ch) for ch in self.channels
        )

    @property
    def parameters(self):
        return dict(
            (p.name, p.value) for p in self.project_leaderboard_entry_dto.parameters
        )

    @property
    def properties(self):
        return dict(
            (p.key, p.value) for p in self.project_leaderboard_entry_dto.properties
        )

    @property
    def tags(self):
        return self.project_leaderboard_entry_dto.tags


class Point(object):
    def __init__(self, point_dto):
        self.point_dto = point_dto

    @property
    def x(self):
        return self.point_dto.x

    @property
    def numeric_y(self):
        return self.point_dto.y.numericValue


class Points(object):
    def __init__(self, point_dtos):
        self.point_dtos = point_dtos

    @property
    def xs(self):
        return [p.x for p in self.point_dtos]

    @property
    def numeric_ys(self):
        return [p.y.numericValue for p in self.point_dtos]



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from typing import Union, Iterable, Optional

from neptune.new.attributes.attribute import Attribute
from neptune.new.internal.operation import UploadFileSet, DeleteFiles
from neptune.new.internal.utils import verify_type, verify_collection_type
from neptune.new.types.file_set import FileSet as FileSetVal


# pylint: disable=protected-access


class FileSet(Attribute):

    def assign(self, value: Union[FileSetVal, str, Iterable[str]], wait: bool = False) -> None:
        verify_type("value", value, (FileSetVal, str, Iterable))
        if isinstance(value, FileSetVal):
            value = value.file_globs
        elif isinstance(value, str):
            value = [value]
        else:
            verify_collection_type("value", value, str)
        self._enqueue_upload_operation(value, reset=True, wait=wait)

    def upload_files(self, globs: Union[str, Iterable[str]], wait: bool = False) -> None:
        if isinstance(globs, str):
            globs = [globs]
        else:
            verify_collection_type("globs", globs, str)
        self._enqueue_upload_operation(globs, reset=False, wait=wait)

    def delete_files(self, paths: Union[str, Iterable[str]], wait: bool = False) -> None:
        if isinstance(paths, str):
            paths = [paths]
        else:
            verify_collection_type("paths", paths, str)
        with self._run.lock():
            self._enqueue_operation(DeleteFiles(self._path, set(paths)), wait)

    def _enqueue_upload_operation(self, globs: Iterable[str], reset: bool, wait: bool):
        with self._run.lock():
            abs_file_globs = list(os.path.abspath(file_glob) for file_glob in globs)
            self._enqueue_operation(UploadFileSet(self._path, abs_file_globs, reset=reset), wait)

    def download(self, destination: Optional[str] = None) -> None:
        verify_type("destination", destination, (str, type(None)))
        self._backend.download_file_set(self._run_id, self._path, destination)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Iterable, Union
import typing

from neptune.new.internal.utils import verify_type, verify_collection_type, is_collection

from neptune.new.internal.operation import AddStrings, RemoveStrings, ClearStringSet
from neptune.new.types.sets.string_set import StringSet as StringSetVal
from neptune.new.attributes.sets.set import Set


class StringSet(Set):

    def assign(self, value: StringSetVal, wait: bool = False):
        verify_type("value", value, StringSetVal)
        with self._run.lock():
            if not value.values:
                self._enqueue_operation(ClearStringSet(self._path), wait=wait)
            else:
                self._enqueue_operation(ClearStringSet(self._path), wait=False)
                self._enqueue_operation(AddStrings(self._path, value.values), wait=wait)

    def add(self, values: Union[str, Iterable[str]], wait: bool = False):
        values = self._to_proper_value_type(values)
        with self._run.lock():
            self._enqueue_operation(AddStrings(self._path, set(values)), wait)

    def remove(self, values: Union[str, Iterable[str]], wait: bool = False):
        values = self._to_proper_value_type(values)
        with self._run.lock():
            self._enqueue_operation(RemoveStrings(self._path, set(values)), wait)

    def clear(self, wait: bool = False):
        with self._run.lock():
            self._enqueue_operation(ClearStringSet(self._path), wait)

    def fetch(self) -> typing.Set[str]:
        # pylint: disable=protected-access
        val = self._backend.get_string_set_attribute(self._run_id, self._path)
        return val.values

    @staticmethod
    def _to_proper_value_type(values: Union[str, Iterable[str]]) -> Iterable[str]:
        if is_collection(values):
            verify_collection_type("values", values, str)
            return list(values)
        else:
            verify_type("values", values, str)
            return [values]



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest

from neptune.management.internal.utils import normalize_project_name
from neptune.management.exceptions import InvalidProjectName, ConflictingWorkspaceName, MissingWorkspaceName


class TestManagementUtils(unittest.TestCase):
    def test_normalize_project_name(self):
        self.assertEqual('jackie/sandbox', normalize_project_name(name='jackie/sandbox'))
        self.assertEqual('jackie/sandbox', normalize_project_name(name='sandbox', workspace='jackie'))
        self.assertEqual('jackie/sandbox', normalize_project_name(name='jackie/sandbox', workspace='jackie'))

        with self.assertRaises(InvalidProjectName):
            normalize_project_name(name='nothing/else/matters')

        with self.assertRaises(MissingWorkspaceName):
            normalize_project_name(name='sandbox')

        with self.assertRaises(ConflictingWorkspaceName):
            normalize_project_name(name='jackie/sandbox', workspace='john')



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest
import pathlib
from urllib.parse import urlparse

from neptune.new.internal.artifacts.types import ArtifactDriversMap, ArtifactDriver, ArtifactFileData
from neptune.new.exceptions import NeptuneUnhandledArtifactSchemeException, NeptuneUnhandledArtifactTypeException


class TestArtifactDriversMap(unittest.TestCase):
    # pylint:disable=protected-access

    def setUp(self):
        self._impl_backup = ArtifactDriversMap._implementations
        ArtifactDriversMap._implementations = []

        class TestArtifactDriver(ArtifactDriver):
            @staticmethod
            def get_type():
                return "test"

            @classmethod
            def matches(cls, path: str) -> bool:
                return urlparse(path).scheme == 'test'

            @classmethod
            def get_tracked_files(cls, path, destination=None):
                return []

            @classmethod
            def download_file(cls, destination: pathlib.Path, file_definition: ArtifactFileData):
                pass

        self.test_driver_instance = TestArtifactDriver
        ArtifactDriversMap._implementations = [self.test_driver_instance]

    def tearDown(self):
        ArtifactDriversMap._implementations = self._impl_backup

    def test_driver_autoregister(self):
        class PkArtifactDriver(ArtifactDriver):
            @staticmethod
            def get_type() -> str:
                return "PK"

            @classmethod
            def matches(cls, path: str) -> bool:
                return urlparse(path).scheme == 'pk'

            @classmethod
            def get_tracked_files(cls, path, destination=None):
                return []

            @classmethod
            def download_file(cls, destination: pathlib.Path, file_definition: ArtifactFileData):
                pass

        self.assertIn(
            PkArtifactDriver,
            ArtifactDriversMap._implementations,
            "PkArtifactDriver not registered with subclass logic"
        )

    def test_match_by_path(self):
        driver_instance = ArtifactDriversMap.match_path("test://path/to/file")

        self.assertEqual(driver_instance, self.test_driver_instance)

    def test_unmatched_path_raises_exception(self):
        with self.assertRaises(NeptuneUnhandledArtifactSchemeException):
            ArtifactDriversMap.match_path("test2://path/to/file")

    def test_match_by_type(self):
        driver_instance = ArtifactDriversMap.match_type("test")

        self.assertEqual(driver_instance, self.test_driver_instance)

    def test_unmatched_type_raises_exception(self):
        with self.assertRaises(NeptuneUnhandledArtifactTypeException):
            ArtifactDriversMap.match_type("test2")



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Remember to set environment values:
* NEPTUNE_API_TOKEN
* NEPTUNE_PROJECT
"""
import sys
from datetime import datetime

from requests import HTTPError

import neptune
from alpha_integration_dev.common_client_code import ClientFeatures
from neptune.exceptions import (
    DeleteArtifactUnsupportedInAlphaException,
    DownloadArtifactUnsupportedException,
    DownloadArtifactsUnsupportedException,
    DownloadSourcesException,
    FileNotFound,
)
from neptune.internal.api_clients.hosted_api_clients.hosted_alpha_leaderboard_api_client \
    import HostedAlphaLeaderboardApiClient


def get_api_version(exp):
    backend = exp._backend
    if isinstance(backend, HostedAlphaLeaderboardApiClient):
        return 2
    return 1


class OldClientFeatures(ClientFeatures):
    def __init__(self):
        super().__init__()
        neptune.init()
        neptune.create_experiment(
            name='const project name',
            description='exp description',
            params=self.params,
            properties=self.properties,
            tags=['initial tag 1', 'initial tag 2'],
            abort_callback=None,
            run_monitoring_thread=False,
            hostname='hostname value',
            # notebook_id='test1',  # TODO: Error 500 when wrong value
            upload_source_files='alpha_integration_dev/*.py',
        )

        exp = neptune.get_experiment()

        self._api_version = get_api_version(exp)

        properties = exp.get_properties()
        assert properties['init_text_property'] == 'some text'
        assert properties['init_number property'] == '42'
        assert properties['init_list'] == '[1, 2, 3]'

        assert set(exp.get_tags()) == {'initial tag 1', 'initial tag 2'}

        # download sources
        if self._api_version == 1:
            # old domain

            with self.with_check_if_file_appears('old_client.py.zip'):
                exp.download_sources('alpha_integration_dev/old_client.py')
            with self.with_check_if_file_appears('alpha_integration_dev.zip'):
                exp.download_sources('alpha_integration_dev')

            with self.with_assert_raises(FileNotFound):
                exp.download_sources('non_existing')
        else:
            # new api

            with self.with_check_if_file_appears('files.zip'):
                exp.download_sources()
            with self.with_assert_raises(DownloadSourcesException):
                exp.download_sources('whatever')
            with self.with_check_if_file_appears('file_set_sources/files.zip'):
                exp.download_sources(destination_dir='file_set_sources')

    def modify_tags(self):
        neptune.append_tags('tag1')
        neptune.append_tag(['tag2_to_remove', 'tag3'])
        neptune.remove_tag('tag2_to_remove')
        neptune.remove_tag('tag4_remove_non_existing')

        exp = neptune.get_experiment()
        assert set(exp.get_tags()) == {'initial tag 1', 'initial tag 2', 'tag1', 'tag3'}

    def modify_properties(self):
        neptune.set_property('prop', 'some text')
        neptune.set_property('prop_number', 42)
        neptune.set_property('nested/prop', 42)
        neptune.set_property('prop_to_del', 42)
        neptune.set_property('prop_list', [1, 2, 3])
        with open(self.text_file_path, mode='r') as f:
            neptune.set_property('prop_IO', f)
        neptune.set_property('prop_datetime', datetime.now())
        neptune.remove_property('prop_to_del')

        exp = neptune.get_experiment()
        properties = exp.get_properties()
        assert properties['prop'] == 'some text'
        assert properties['prop_number'] == '42'
        assert properties['nested/prop'] == '42'
        assert properties['prop_list'] == '[1, 2, 3]'
        assert 'prop_to_del' not in properties
        assert properties['prop_IO'] == "<_io.TextIOWrapper name='alpha_integration_dev/data/text.txt'" \
                                        " mode='r' encoding='UTF-8'>"
        print(f'Properties: {properties}')

    def log_std(self):
        print('stdout text1')
        print('stdout text2')
        print('stderr text1', file=sys.stderr)
        print('stderr text2', file=sys.stderr)

    def log_series(self):
        # floats
        neptune.log_metric('m1', 1)
        neptune.log_metric('m1', 2)
        neptune.log_metric('m1', 3)
        neptune.log_metric('m1', 2)
        neptune.log_metric('nested/m1', 1)

        # texts
        neptune.log_text('m2', 'a')
        neptune.log_text('m2', 'b')
        neptune.log_text('m2', 'c')

        # images
        # `image_name` and `description` will be lost
        neptune.log_image('g_img', self.img_path, image_name='name', description='desc')
        neptune.log_image('g_img', self.img_path)

        # see what we've logged
        logs = neptune.get_experiment().get_logs()
        print(f'Logs: {logs}')

    def handle_files_and_images(self):
        # image
        # `image_name` and `description` will be lost (`send_image` the same as `log_image`)
        neptune.send_image('image', self.img_path, name='name', description='desc')

        # artifact with default dest
        neptune.send_artifact(self.text_file_path)
        exp = neptune.get_experiment()
        with self.with_check_if_file_appears('text.txt'):
            exp.download_artifact('text.txt')
        with self.with_check_if_file_appears('custom_dest/text.txt'):
            exp.download_artifact('text.txt', 'custom_dest')

        # artifact with custom dest
        neptune.send_artifact(self.text_file_path, destination='something.txt')
        exp = neptune.get_experiment()
        with self.with_check_if_file_appears('something.txt'):
            exp.download_artifact('something.txt')
        with self.with_check_if_file_appears('custom_dest/something.txt'):
            exp.download_artifact('something.txt', 'custom_dest')

        # destination dirs
        neptune.log_artifact(self.text_file_path, destination='dir/text file artifact')
        neptune.log_artifact(self.text_file_path, destination='dir/artifact_to_delete')

        # deleting
        neptune.delete_artifacts('dir/artifact_to_delete')

        # streams
        with open(self.text_file_path, mode='r') as f:
            neptune.send_artifact(f, destination='file stream.txt')

    def handle_directories(self):
        exp = neptune.get_experiment()

        # download_artifacts
        neptune.send_artifact(self.data_dir)
        if self._api_version == 1:
            with self.with_check_if_file_appears('output.zip'):
                exp.download_artifacts()
        else:
            with self.with_assert_raises(DownloadArtifactsUnsupportedException):
                exp.download_artifacts()

        # create some nested artifacts
        neptune.log_artifact(self.img_path, destination='main dir/sub dir/art1')
        neptune.log_artifact(self.img_path, destination='main dir/sub dir/art2')
        neptune.log_artifact(self.img_path, destination='main dir/sub dir/art3')

        # downloading artifact - download_artifact
        # non existing artifact
        if self._api_version == 1:
            with self.with_assert_raises(FileNotFound):
                exp.download_artifact('main dir/sub dir/art100')
        else:
            with self.with_assert_raises(DownloadArtifactUnsupportedException):
                exp.download_artifact('main dir/sub dir/art100')
        # artifact directories
        if self._api_version == 1:
            with self.with_assert_raises(HTTPError):
                exp.download_artifact('main dir/sub dir')
        else:
            with self.with_assert_raises(DownloadArtifactUnsupportedException):
                exp.download_artifact('main dir/sub dir')

        # deleting artifacts
        neptune.delete_artifacts('main dir/sub dir/art1')

        # delete non existing artifact
        if self._api_version == 1:
            neptune.delete_artifacts('main dir/sub dir/art100')
        else:
            with self.with_assert_raises(DeleteArtifactUnsupportedInAlphaException):
                neptune.delete_artifacts('main dir/sub dir/art100')

        # delete dir
        if self._api_version == 1:
            neptune.delete_artifacts('main dir/sub dir')
        else:
            with self.with_assert_raises(DeleteArtifactUnsupportedInAlphaException):
                neptune.delete_artifacts('main dir/sub dir')

    def finalize(self):
        pass


if __name__ == '__main__':
    OldClientFeatures().run()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access

from mock import MagicMock, patch

from neptune.new.attributes.series.float_series import FloatSeries

from tests.neptune.new.attributes.test_attribute_base import TestAttributeBase


@patch("time.time", new=TestAttributeBase._now)
class TestFloatSeries(TestAttributeBase):

    def test_assign_type_error(self):
        values = [["text"], 55, "string", None]
        for value in values:
            with self.assertRaises(Exception):
                FloatSeries(MagicMock(), MagicMock()).assign(value)

    def test_log_type_error(self):
        values = [["text"], [[]], [5, ""], "string", None]
        for value in values:
            with self.assertRaises(Exception):
                FloatSeries(MagicMock(), MagicMock()).log(value)

    def test_get(self):
        exp, path = self._create_run(), self._random_path()
        var = FloatSeries(exp, path)
        var.log(5)
        var.log(34)
        self.assertEqual(34, var.fetch_last())



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import logging
import os
import threading
from glob import glob
from pathlib import Path
from typing import TypeVar, List, Callable, Optional, Tuple

from neptune.new.exceptions import MalformedOperation
from neptune.new.internal.containers.storage_queue import StorageQueue
from neptune.new.internal.utils.json_file_splitter import JsonFileSplitter
from neptune.new.internal.utils.sync_offset_file import SyncOffsetFile

T = TypeVar('T')

_logger = logging.getLogger(__name__)


class DiskQueue(StorageQueue[T]):

    # NOTICE: This class is thread-safe as long as there is only one consumer and one producer.

    def __init__(
            self,
            dir_path: Path,
            to_dict: Callable[[T], dict],
            from_dict: Callable[[dict], T],
            lock: threading.RLock,
            max_file_size: int = 64 * 1024**2):
        self._dir_path = dir_path.resolve()
        self._to_dict = to_dict
        self._from_dict = from_dict
        self._max_file_size = max_file_size

        try:
            os.makedirs(self._dir_path)
        except FileExistsError:
            pass

        self._last_ack_file = SyncOffsetFile(dir_path / "last_ack_version", default=0)
        self._last_put_file = SyncOffsetFile(dir_path / "last_put_version", default=0)

        self._read_file_version, self._write_file_version = self._get_first_and_last_log_file_version()
        self._writer = open(self._get_log_file(self._write_file_version), "a")
        self._reader = JsonFileSplitter(self._get_log_file(self._read_file_version))
        self._file_size = 0
        self._should_skip_to_ack = True

        self._empty_cond = threading.Condition(lock)

    def put(self, obj: T) -> int:
        version = self._last_put_file.read_local() + 1
        _json = json.dumps(self._serialize(obj, version))
        if self._file_size + len(_json) > self._max_file_size:
            self._writer.flush()
            self._writer.close()
            self._writer = open(self._get_log_file(version), "a")
            self._file_size = 0
            self._write_file_version = version
        self._writer.write(_json + "\n")
        self._last_put_file.write(version)
        self._file_size += len(_json) + 1
        return version

    def get(self) -> Tuple[Optional[T], int]:
        if self._should_skip_to_ack:
            return self._skip_and_get()
        else:
            return self._get()

    def _skip_and_get(self) -> Tuple[Optional[T], int]:
        ack_version = self._last_ack_file.read_local()
        ver = -1
        while True:
            obj, next_ver = self._get()
            if obj is None:
                return None, ver
            ver = next_ver
            if ver > ack_version:
                self._should_skip_to_ack = False
                if ver > ack_version + 1:
                    _logger.warning("Possible data loss. Last acknowledged operation version: %d, next: %d",
                                    ack_version, ver)
                return obj, ver

    def _get(self) -> Tuple[Optional[T], int]:
        _json = self._reader.get()
        if not _json:
            if self._read_file_version >= self._write_file_version:
                return None, -1
            self._reader.close()
            self._read_file_version = self._next_log_file_version(self._read_file_version)
            self._reader = JsonFileSplitter(self._get_log_file(self._read_file_version))
            # It is safe. Max recursion level is 2.
            return self._get()
        try:
            return self._deserialize(_json)
        except Exception as e:
            raise MalformedOperation from e

    def get_batch(self, size: int) -> Tuple[List[T], int]:
        first, ver = self.get()
        if not first:
            return [], ver
        ret = [first]
        for _ in range(0, size - 1):
            obj, next_ver = self._get()
            if not obj:
                break
            ver = next_ver
            ret.append(obj)
        return ret, ver

    def flush(self):
        self._writer.flush()
        self._last_ack_file.flush()
        self._last_put_file.flush()

    def close(self):
        self._reader.close()
        self._writer.close()
        self._last_ack_file.close()
        self._last_put_file.close()

    def wait_for_empty(self, seconds: Optional[float] = None) -> bool:
        with self._empty_cond:
            return self._empty_cond.wait_for(self.is_empty, timeout=seconds)

    def ack(self, version: int) -> None:
        self._last_ack_file.write(version)

        log_versions = self._get_all_log_file_versions()
        for i in range(0, len(log_versions) - 1):
            if log_versions[i + 1] <= version:
                filename = self._get_log_file(log_versions[i])
                try:
                    os.remove(filename)
                except FileNotFoundError:
                    # not really a problem
                    pass
                except Exception:
                    _logger.exception("Cannot remove queue file %s", filename)
            else:
                break

        with self._empty_cond:
            if self.is_empty():
                self._empty_cond.notifyAll()

    def is_empty(self) -> bool:
        return self.size() == 0

    def size(self) -> int:
        return self._last_put_file.read_local() - self._last_ack_file.read_local()

    def _get_log_file(self, index: int) -> str:
        return "{}/data-{}.log".format(self._dir_path, index)

    def _get_all_log_file_versions(self):
        log_files = glob("{}/data-*.log".format(self._dir_path))
        if not log_files:
            return 1, 1
        return sorted([int(file[len(str(self._dir_path)) + 6:-4]) for file in log_files])

    def _get_first_and_last_log_file_version(self) -> (int, int):
        log_versions = self._get_all_log_file_versions()
        return min(log_versions), max(log_versions)

    def _next_log_file_version(self, version: int) -> int:
        log_versions = self._get_all_log_file_versions()
        for i, val in enumerate(log_versions):
            if val == version:
                return log_versions[i + 1]
        raise ValueError("Missing log file with version > {}".format(version))

    def _serialize(self, obj: T, version: int) -> dict:
        return {
            "obj": self._to_dict(obj),
            "version": version
        }

    def _deserialize(self, data: dict) -> Tuple[T, int]:
        return self._from_dict(data["obj"]), data["version"]



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Type, List, Optional, Dict, Tuple

from bravado.client import SwaggerClient
from bravado.exception import HTTPNotFound

from neptune.new.internal.artifacts.types import ArtifactDriversMap, ArtifactDriver, ArtifactFileData
from neptune.new.internal.artifacts.file_hasher import FileHasher
from neptune.new.exceptions import (
    ArtifactUploadingError,
    ArtifactNotFoundException,
    NeptuneEmptyLocationException
)
from neptune.new.internal.backends.api_model import ArtifactModel
from neptune.new.internal.backends.utils import with_api_exceptions_handler
from neptune.new.internal.operation import Operation, AssignArtifact


def _compute_artifact_size(artifact_file_list: List[ArtifactFileData]):
    artifact_size = 0
    for artifact_file in artifact_file_list:
        if artifact_file.size is None:
            # whole artifact's size is undefined in this case
            return None
        artifact_size += artifact_file.size
    return artifact_size


def track_to_new_artifact(
        swagger_client: SwaggerClient,
        project_id: str,
        path: List[str],
        parent_identifier: str,
        entries: List[Tuple[str, Optional[str]]],
        default_request_params: Dict
) -> Optional[Operation]:
    files: List[ArtifactFileData] = _extract_file_list(path, entries)

    if not files:
        raise ArtifactUploadingError("Uploading an empty Artifact")

    artifact_hash = _compute_artifact_hash(files)
    artifact = create_new_artifact(
        swagger_client=swagger_client,
        project_id=project_id,
        artifact_hash=artifact_hash,
        parent_identifier=parent_identifier,
        size=_compute_artifact_size(files),
        default_request_params=default_request_params
    )

    if not artifact.received_metadata:
        upload_artifact_files_metadata(
            swagger_client=swagger_client,
            project_id=project_id,
            artifact_hash=artifact_hash,
            files=files,
            default_request_params=default_request_params
        )

    return AssignArtifact(path=path, hash=artifact_hash)


def track_to_existing_artifact(
        swagger_client: SwaggerClient,
        project_id: str,
        path: List[str],
        artifact_hash: str,
        parent_identifier: str,
        entries: List[Tuple[str, Optional[str]]],
        default_request_params: Dict
) -> Optional[Operation]:
    files: List[ArtifactFileData] = _extract_file_list(path, entries)

    if not files:
        raise ArtifactUploadingError("Uploading an empty Artifact")

    artifact = create_artifact_version(
        swagger_client=swagger_client,
        project_id=project_id,
        artifact_hash=artifact_hash,
        parent_identifier=parent_identifier,
        files=files,
        default_request_params=default_request_params
    )

    return AssignArtifact(path=path, hash=artifact.hash)


def _compute_artifact_hash(files: List[ArtifactFileData]) -> str:
    return FileHasher.get_artifact_hash(files)


def _extract_file_list(path: List[str], entries: List[Tuple[str, Optional[str]]]) -> List[ArtifactFileData]:
    files: List[ArtifactFileData] = list()

    for entry_path, entry_destination in entries:
        driver: Type[ArtifactDriver] = ArtifactDriversMap.match_path(entry_path)
        artifact_files = driver.get_tracked_files(path=entry_path, destination=entry_destination)

        if len(artifact_files) == 0:
            raise NeptuneEmptyLocationException(location=entry_path, namespace='/'.join(path))

        files.extend(artifact_files)

    return files


@with_api_exceptions_handler
def create_new_artifact(
        swagger_client: SwaggerClient,
        project_id: str,
        artifact_hash: str,
        parent_identifier: str,
        size: int,
        default_request_params: Dict
) -> ArtifactModel:
    params = {
        'projectIdentifier': project_id,
        'hash': artifact_hash,
        'size': size,
        'parentIdentifier': parent_identifier,
        **default_request_params
    }
    try:
        result = swagger_client.api.createNewArtifact(**params).response().result
        return ArtifactModel(
            hash=result.artifactHash,
            received_metadata=result.receivedMetadata,
            size=result.size
        )
    except HTTPNotFound:
        raise ArtifactNotFoundException(artifact_hash)


@with_api_exceptions_handler
def upload_artifact_files_metadata(
        swagger_client: SwaggerClient,
        project_id: str,
        artifact_hash: str,
        files: List[ArtifactFileData],
        default_request_params: Dict
) -> ArtifactModel:
    params = {
        'projectIdentifier': project_id,
        'hash': artifact_hash,
        'artifactFilesDTO': {
            'files': [
                ArtifactFileData.to_dto(a) for a in files
            ]
        },
        **default_request_params
    }
    try:
        result = swagger_client.api.uploadArtifactFilesMetadata(**params).response().result
        return ArtifactModel(
            hash=result.artifactHash,
            size=result.size,
            received_metadata=result.receivedMetadata
        )
    except HTTPNotFound:
        raise ArtifactNotFoundException(artifact_hash)


@with_api_exceptions_handler
def create_artifact_version(
        swagger_client: SwaggerClient,
        project_id: str,
        artifact_hash: str,
        parent_identifier: str,
        files: List[ArtifactFileData],
        default_request_params: Dict
) -> ArtifactModel:
    params = {
        'projectIdentifier': project_id,
        'hash': artifact_hash,
        'parentIdentifier': parent_identifier,
        'artifactFilesDTO': {
            'files': [
                ArtifactFileData.to_dto(a) for a in files
            ]
        },
        **default_request_params
    }
    try:
        result = swagger_client.api.createArtifactVersion(**params).response().result
        return ArtifactModel(
            hash=result.artifactHash,
            size=result.size,
            received_metadata=result.receivedMetadata
        )
    except HTTPNotFound:
        raise ArtifactNotFoundException(artifact_hash)



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import List, TYPE_CHECKING

from neptune.new.attributes import (
    Artifact, Boolean, Datetime, File, FileSeries, FileSet, Float, FloatSeries, GitRef, Integer, NotebookRef, RunState,
    String, StringSeries, StringSet
)
from neptune.new.exceptions import InternalClientError
from neptune.new.internal.backends.api_model import AttributeType

if TYPE_CHECKING:
    from neptune.new import Run
    from neptune.new.attributes.attribute import Attribute

_attribute_type_to_attr_class_map = {
    AttributeType.FLOAT: Float,
    AttributeType.INT: Integer,
    AttributeType.BOOL: Boolean,
    AttributeType.STRING: String,
    AttributeType.DATETIME: Datetime,
    AttributeType.FILE: File,
    AttributeType.FILE_SET: FileSet,
    AttributeType.FLOAT_SERIES: FloatSeries,
    AttributeType.STRING_SERIES: StringSeries,
    AttributeType.IMAGE_SERIES: FileSeries,
    AttributeType.STRING_SET: StringSet,
    AttributeType.GIT_REF: GitRef,
    AttributeType.RUN_STATE: RunState,
    AttributeType.NOTEBOOK_REF: NotebookRef,
    AttributeType.ARTIFACT: Artifact,
}


def create_attribute_from_type(attribute_type: AttributeType, run: 'Run', path: List[str]) -> 'Attribute':
    try:
        return _attribute_type_to_attr_class_map[attribute_type](run, path)
    except KeyError:
        raise InternalClientError(f"Unexpected type: {attribute_type}")



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from typing import Optional

from neptune.new.internal.utils import verify_type, base64_encode

from neptune.new.internal.operation import UploadFile, UploadFileContent
from neptune.new.types.atoms.file import File as FileVal
from neptune.new.attributes.atoms.atom import Atom

# pylint: disable=protected-access


class File(Atom):

    def assign(self, value: FileVal, wait: bool = False) -> None:
        verify_type("value", value, FileVal)

        if value.path is not None:
            operation = UploadFile(self._path, ext=value.extension, file_path=os.path.abspath(value.path))
        elif value.content is not None:
            operation = UploadFileContent(self._path, ext=value.extension, file_content=base64_encode(value.content))
        else:
            raise ValueError("File path and content are None")

        with self._run.lock():
            self._enqueue_operation(operation, wait)

    def upload(self, value, wait: bool = False) -> None:
        self.assign(FileVal.create_from(value), wait)

    def download(self, destination: Optional[str] = None) -> None:
        verify_type("destination", destination, (str, type(None)))
        self._backend.download_file(self._run_id, self._path, destination)

    def fetch_extension(self):
        # pylint: disable=protected-access
        val = self._backend.get_file_attribute(self._run_id, self._path)
        return val.ext



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import base64
from io import IOBase, BytesIO
import logging
import os
from glob import glob
from typing import Union, TypeVar, Iterable, List, Set, Optional, Mapping

from neptune.new.internal.utils import limits


T = TypeVar('T')

_logger = logging.getLogger(__name__)


def replace_patch_version(version: str):
    return version[:version.index(".", version.index(".") + 1)] + ".0"


def verify_type(var_name: str, var, expected_type: Union[type, tuple]):
    try:
        if isinstance(expected_type, tuple):
            type_name = " or ".join(get_type_name(t) for t in expected_type)
        else:
            type_name = get_type_name(expected_type)
    except Exception as e:
        # Just to be sure that nothing weird will be raised here
        raise TypeError("Incorrect type of {}".format(var_name)) from e

    if not isinstance(var, expected_type):
        raise TypeError("{} must be a {} (was {})".format(var_name, type_name, type(var)))

    if isinstance(var, IOBase) and not hasattr(var, 'read'):
        raise TypeError("{} is a stream, which does not implement read method".format(var_name))


def is_stream(var):
    return isinstance(var, IOBase) and hasattr(var, 'read')


def is_bool(var):
    return isinstance(var, bool)


def is_int(var):
    return isinstance(var, int)


def is_float(var):
    return isinstance(var, (float, int))


def is_string(var):
    return isinstance(var, str)


def is_float_like(var):
    try:
        _ = float(var)
        return True
    except (ValueError, TypeError):
        return False


def is_dict_like(var):
    return isinstance(var, (dict, Mapping))


def is_string_like(var):
    try:
        _ = str(var)
        return True
    except ValueError:
        return False


def get_type_name(_type: Union[type, tuple]):
    return _type.__name__ if hasattr(_type, '__name__') else str(_type)


def verify_collection_type(var_name: str, var, expected_type: Union[type, tuple]):
    verify_type(var_name, var, (list, set, tuple))
    for value in var:
        verify_type("elements of collection '{}'".format(var_name), value, expected_type)


def is_collection(var) -> bool:
    return isinstance(var, (list, set, tuple))


def base64_encode(data: bytes) -> str:
    return base64.b64encode(data).decode('utf-8')


def base64_decode(data: str) -> bytes:
    return base64.b64decode(data.encode('utf-8'))


def get_absolute_paths(file_globs: Iterable[str]) -> List[str]:
    expanded_paths: Set[str] = set()
    for file_glob in file_globs:
        expanded_paths |= set(glob(file_glob, recursive=True))
    return list(os.path.abspath(expanded_file) for expanded_file in expanded_paths)


def get_common_root(absolute_paths: List[str]) -> Optional[str]:
    try:
        common_root = os.path.commonpath(absolute_paths)
        if os.path.isfile(common_root):
            common_root = os.path.dirname(common_root)
        if common_root.startswith(os.getcwd() + os.sep):
            common_root = os.getcwd()
        return common_root
    except ValueError:
        return None


def get_stream_content(stream: IOBase, seek: Optional[int] = None) -> (Optional[str], str):
    if seek is not None and stream.seekable():
        stream.seek(seek)

    content = stream.read(limits.STREAM_SIZE_LIMIT_BYTES + 1)
    default_ext = "txt" if isinstance(content, str) else "bin"
    if isinstance(content, str):
        content = content.encode('utf-8')

    if limits.stream_size_exceeds_limit(len(content)):
        return None, default_ext

    while True:
        chunk = stream.read(1024 * 1024)
        if chunk is None:
            continue
        elif not chunk:
            break
        else:
            if not isinstance(content, BytesIO):
                content = BytesIO(content)
                content.seek(0, 2)
            if isinstance(chunk, str):
                chunk = chunk.encode('utf-8')
            if limits.stream_size_exceeds_limit(content.tell() + len(chunk)):
                return None, default_ext
            content.write(chunk)
    if isinstance(content, BytesIO):
        content = content.getvalue()

    return content, default_ext


def is_ipython() -> bool:
    try:
        # pylint:disable=bad-option-value,import-outside-toplevel
        import IPython
        ipython = IPython.core.getipython.get_ipython()
        return ipython is not None
    except ImportError:
        return False



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging

_logger = logging.getLogger(__name__)


_IMAGE_SIZE_LIMIT_MB = 15
_IN_MEMORY_SIZE_LIMIT_MB = 32
_STREAM_SIZE_LIMIT_MB = 32

BYTES_IN_MB = 1024 * 1024

STREAM_SIZE_LIMIT_BYTES = _STREAM_SIZE_LIMIT_MB * BYTES_IN_MB


def image_size_exceeds_limit(content_size):
    if content_size > _IMAGE_SIZE_LIMIT_MB * BYTES_IN_MB:
        _logger.warning('You are attempting to create an image that is %.2fMB large. '
                        'Neptune supports logging images smaller than %dMB. '
                        'Resize or increase compression of this image',
                        content_size / BYTES_IN_MB,
                        _IMAGE_SIZE_LIMIT_MB)
        return True
    return False


def file_size_exceeds_limit(content_size):
    if content_size > _IN_MEMORY_SIZE_LIMIT_MB * BYTES_IN_MB:
        _logger.warning('You are attempting to create an in-memory file that is %.1fMB large. '
                        'Neptune supports logging in-memory file objects smaller than %dMB. '
                        'Resize or increase compression of this object',
                        content_size / BYTES_IN_MB,
                        _IN_MEMORY_SIZE_LIMIT_MB)
        return True
    return False


def stream_size_exceeds_limit(content_size):
    if content_size > _STREAM_SIZE_LIMIT_MB * BYTES_IN_MB:
        _logger.warning('Your stream is larger than %dMB. '
                        'Neptune supports saving files from streams smaller than %dMB.',
                        _STREAM_SIZE_LIMIT_MB,
                        _STREAM_SIZE_LIMIT_MB)
        return True
    return False



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import os
import sys
import threading
from time import time, monotonic
from typing import Optional, List

import click

from neptune.new.internal.containers.storage_queue import StorageQueue
from neptune.new.internal.backends.neptune_backend import NeptuneBackend
from neptune.new.internal.operation import Operation
from neptune.new.internal.operation_processors.operation_processor import OperationProcessor
from neptune.new.internal.threading.daemon import Daemon

# pylint: disable=protected-access

_logger = logging.getLogger(__name__)


class AsyncOperationProcessor(OperationProcessor):
    STOP_QUEUE_STATUS_UPDATE_FREQ_SECONDS = 30
    STOP_QUEUE_MAX_TIME_NO_CONNECTION_SECONDS = 300

    def __init__(self,
                 run_id: str,
                 queue: StorageQueue[Operation],
                 backend: NeptuneBackend,
                 lock: threading.RLock,
                 sleep_time: float = 5,
                 batch_size: int = 1000):
        self._run_id = run_id
        self._queue = queue
        self._backend = backend
        self._batch_size = batch_size
        self._last_version = 0
        self._consumed_version = 0
        self._consumer = self.ConsumerThread(self, sleep_time, batch_size)
        self._drop_operations = False

        # Caller is responsible for taking this lock
        self._waiting_cond = threading.Condition(lock=lock)

        if sys.version_info >= (3, 7):
            try:
                # pylint: disable=no-member
                os.register_at_fork(after_in_child=self._handle_fork_in_child)
            except AttributeError:
                pass

    def _handle_fork_in_child(self):
        self._drop_operations = True

    def enqueue_operation(self, op: Operation, wait: bool) -> None:
        if self._drop_operations:
            return
        self._last_version = self._queue.put(op)
        if self._queue.size() > self._batch_size / 2:
            self._consumer.wake_up()
        if wait:
            self.wait()

    def wait(self):
        self.flush()
        waiting_for_version = self._last_version
        self._consumer.wake_up()
        self._waiting_cond.wait_for(lambda: self._consumed_version >= waiting_for_version)

    def flush(self):
        self._queue.flush()

    def start(self):
        self._consumer.start()

    def _wait_for_queue_empty(self, initial_queue_size: int, seconds: Optional[float]):
        waiting_start = monotonic()
        time_elapsed = 0
        max_reconnect_wait_time = self.STOP_QUEUE_MAX_TIME_NO_CONNECTION_SECONDS if seconds is None else seconds
        if initial_queue_size > 0:
            if self._consumer.last_backoff_time > 0:
                click.echo(f"We have been experiencing connection interruptions during your run. "
                           f"Neptune client will now try to resume connection and sync data for the next "
                           f"{max_reconnect_wait_time} seconds. "
                           f"You can also kill this process and synchronize your data manually later "
                           f"using `neptune sync` command.",
                           sys.stderr)
            else:
                click.echo(f"Waiting for the remaining {initial_queue_size} operations to synchronize with Neptune. "
                           f"Do not kill this process.",
                           sys.stderr)

        while True:
            if seconds is None:
                wait_time = self.STOP_QUEUE_STATUS_UPDATE_FREQ_SECONDS
            else:
                wait_time = max(
                    min(seconds - time_elapsed, self.STOP_QUEUE_STATUS_UPDATE_FREQ_SECONDS),
                    0
                )
            self._queue.wait_for_empty(wait_time)
            size_remaining = self._queue.size()
            already_synced = initial_queue_size - size_remaining
            already_synced_proc = (already_synced / initial_queue_size) * 100 if initial_queue_size else 100
            if size_remaining == 0:
                click.echo(f"All {initial_queue_size} operations synced, thanks for waiting!")
                return

            time_elapsed = monotonic() - waiting_start
            if self._consumer.last_backoff_time > 0 and time_elapsed >= max_reconnect_wait_time:
                click.echo(
                    f"Failed to reconnect with Neptune in {max_reconnect_wait_time} seconds."
                    f" You have {size_remaining} operations saved on disk that can be manually synced"
                    f" using `neptune sync` command.",
                    sys.stderr)
                return

            if seconds is not None and wait_time == 0:
                click.echo(
                    f"Failed to sync all operations in {seconds} seconds."
                    f" You have {size_remaining} operations saved on disk that can be manually synced"
                    f" using `neptune sync` command.",
                    sys.stderr)
                return

            click.echo(
                f"Still waiting for the remaining {size_remaining} operations "
                f"({already_synced_proc:.2f}% done). Please wait.",
                sys.stderr)

    def stop(self, seconds: Optional[float] = None):
        ts = time()
        self._queue.flush()
        if self._consumer.is_running():
            self._consumer.disable_sleep()
            self._consumer.wake_up()
            self._wait_for_queue_empty(initial_queue_size=self._queue.size(), seconds=seconds)
            self._consumer.interrupt()
        sec_left = None if seconds is None else seconds - (time() - ts)
        self._consumer.join(sec_left)
        self._queue.close()

    class ConsumerThread(Daemon):
        def __init__(self,
                     processor: 'AsyncOperationProcessor',
                     sleep_time: float,
                     batch_size: int):
            super().__init__(sleep_time=sleep_time)
            self._processor = processor
            self._batch_size = batch_size
            self._last_flush = 0

        def work(self) -> None:
            ts = time()
            if ts - self._last_flush >= self._sleep_time:
                self._last_flush = ts
                self._processor._queue.flush()

            while True:
                batch, version = self._processor._queue.get_batch(self._batch_size)
                if not batch:
                    return
                self.process_batch(batch, version)

        @Daemon.ConnectionRetryWrapper(
            kill_message=(
                "Killing Neptune asynchronous thread. All data is safe on disk and can be later"
                " synced manually using `neptune sync` command."
            )
        )
        def process_batch(self, batch: List[Operation], version: int) -> None:
            # TODO: Handle Metadata errors
            result = self._processor._backend.execute_operations(self._processor._run_id, batch)

            with self._processor._waiting_cond:
                self._processor._queue.ack(version)

                for error in result:
                    _logger.error("Error occurred during asynchronous operation processing: %s", error)

                self._processor._consumed_version = version
                self._processor._waiting_cond.notify_all()



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access
import contextlib
import io
import os
import unittest
from typing import Optional
from unittest import mock
from uuid import uuid4

import matplotlib
import pandas
from matplotlib import pyplot
from matplotlib.figure import Figure

import plotly.express as px
import altair as alt
from vega_datasets import data
from bokeh.plotting import figure

from PIL import Image
import numpy

from neptune.new.internal.utils.images import get_image_content, get_html_content
from neptune.utils import IS_WINDOWS

matplotlib.use('agg')


class TestImage(unittest.TestCase):

    TEST_DIR = "/tmp/neptune/{}".format(uuid4())

    def setUp(self):
        if not os.path.exists(self.TEST_DIR):
            os.makedirs(self.TEST_DIR)

    def test_get_image_content_from_pil_image(self):
        # given
        image_array = self._random_image_array()
        expected_image = Image.fromarray(image_array.astype(numpy.uint8))

        # expect
        self.assertEqual(get_image_content(expected_image), self._encode_pil_image(expected_image))

    def test_get_image_content_from_2d_grayscale_array(self):
        # given
        image_array = self._random_image_array(d=None)
        scaled_array = image_array * 255
        expected_image = Image.fromarray(scaled_array.astype(numpy.uint8))

        # expect
        self.assertEqual(get_image_content(image_array), self._encode_pil_image(expected_image))

    def test_get_image_content_from_3d_grayscale_array(self):
        # given
        image_array = numpy.array([
            [[1], [0]],
            [[-3], [4]],
            [[5], [6]]
        ])
        expected_array = numpy.array([
            [1, 0],
            [-3, 4],
            [5, 6]
        ]) * 255
        expected_image = Image.fromarray(expected_array.astype(numpy.uint8))

        # expect
        stderr = io.StringIO()
        with contextlib.redirect_stderr(stderr):
            self.assertEqual(get_image_content(image_array), self._encode_pil_image(expected_image))
        self.assertEqual(
            stderr.getvalue(),
            "The smallest value in the array is -3 and the largest value in the array is 6."
            " To be interpreted as colors correctly values in the array need to be in the [0, 1] range.\n"
        )

    def test_get_image_content_from_rgb_array(self):
        # given
        image_array = self._random_image_array()
        scaled_array = image_array * 255
        expected_image = Image.fromarray(scaled_array.astype(numpy.uint8))

        # expect
        self.assertEqual(get_image_content(image_array), self._encode_pil_image(expected_image))

    def test_get_image_content_from_rgba_array(self):
        # given
        image_array = self._random_image_array(d=4)
        scaled_array = image_array * 255
        expected_image = Image.fromarray(scaled_array.astype(numpy.uint8))

        # expect
        self.assertEqual(get_image_content(image_array), self._encode_pil_image(expected_image))

    def test_get_image_content_from_figure(self):
        # given
        pyplot.plot([1, 2, 3, 4])
        pyplot.ylabel('some interesting numbers')
        fig = pyplot.gcf()

        # expect
        self.assertEqual(get_image_content(fig), self._encode_figure(fig))

    @unittest.skipIf(IS_WINDOWS, "Installing Torch on Windows takes too long")
    def test_get_image_content_from_torch_tensor(self):
        import torch  # pylint: disable=C0415
        # given
        image_tensor = torch.rand(200, 300, 3)  # pylint: disable=no-member
        expected_array = image_tensor.numpy() * 255
        expected_image = Image.fromarray(expected_array.astype(numpy.uint8))

        # expect
        self.assertEqual(get_image_content(image_tensor), self._encode_pil_image(expected_image))

    def test_get_image_content_from_tensorflow_tensor(self):
        import tensorflow as tf  # pylint: disable=C0415
        # given
        image_tensor = tf.random.uniform(shape=[200, 300, 3])  # pylint: disable=E1120 # false positive
        expected_array = image_tensor.numpy() * 255
        expected_image = Image.fromarray(expected_array.astype(numpy.uint8))

        # expect
        self.assertEqual(get_image_content(image_tensor), self._encode_pil_image(expected_image))

    def test_get_html_from_matplotlib_figure(self):
        # given
        fig = pyplot.figure()
        x = [21, 22, 23, 4, 5, 6, 77, 8, 9, 10, 31, 32, 33, 34, 35, 36, 37, 18, 49, 50, 100]
        pyplot.hist(x, bins=5)

        # when
        result = get_html_content(fig)

        # then
        self.assertTrue(result.startswith('<html>\n<head><meta charset="utf-8" />'))

    def test_get_html_from_plotly(self):
        # given
        df = px.data.tips()
        fig = px.histogram(df, x="total_bill", y="tip", color="sex", marginal="rug",
                           hover_data=df.columns)

        # when
        result = get_html_content(fig)

        # then
        self.assertTrue(result.startswith('<html>\n<head><meta charset="utf-8" />'))

    def test_get_html_from_altair(self):
        # given
        source = data.cars()

        chart = alt.Chart(source).mark_circle(size=60).encode(
            x='Horsepower',
            y='Miles_per_Gallon',
            color='Origin',
            tooltip=['Name', 'Origin', 'Horsepower', 'Miles_per_Gallon']
        ).interactive()

        # when
        result = get_html_content(chart)

        # then
        self.assertTrue(result.startswith('<!DOCTYPE html>\n<html>\n<head>\n  <style>'))

    def test_get_html_from_bokeh(self):
        # given
        p = figure(plot_width=400, plot_height=400)
        p.circle(size=20, color="navy", alpha=0.5)

        # when
        result = get_html_content(p)

        # then
        self.assertTrue(result.startswith('\n\n\n\n<!DOCTYPE html>\n<html lang="en">'))

    def test_get_html_from_pandas(self):
        # given
        table = pandas.DataFrame(
            numpy.random.randn(6, 4),
            index=pandas.date_range("20130101", periods=6),
            columns=list("ABCD"))

        # when
        result = get_html_content(table)

        # then
        self.assertTrue(result.startswith(
            '<table border="1" class="dataframe">\n  <thead>\n    <tr style="text-align: right;">'))

    def test_get_oversize_html_from_pandas(self):
        # given
        table = mock.Mock(spec=pandas.DataFrame)
        table.to_html.return_value = 40_000_000 * 'a'

        # when
        with self.assertLogs() as caplog:
            result = get_html_content(table)

        # then
        self.assertIsNone(result)
        self.assertEqual(
            caplog.output, [
                'WARNING:neptune.new.internal.utils.limits:You are attempting to create an in-memory file that'
                ' is 38.1MB large. Neptune supports logging in-memory file objects smaller than 32MB. '
                'Resize or increase compression of this object'
            ]
        )

    @staticmethod
    def _encode_pil_image(image: Image) -> bytes:
        with io.BytesIO() as image_buffer:
            image.save(image_buffer, format='PNG')
            return image_buffer.getvalue()

    @staticmethod
    def _encode_figure(fig: Figure) -> bytes:
        with io.BytesIO() as image_buffer:
            fig.savefig(image_buffer, format='PNG', bbox_inches="tight")
            return image_buffer.getvalue()

    @staticmethod
    def _random_image_array(w=20, h=30, d: Optional[int] = 3):
        if d:
            return numpy.random.rand(w, h, d)
        else:
            return numpy.random.rand(w, h)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from enum import Enum
from typing import List, TypeVar, Callable

from neptune.new.exceptions import MetadataInconsistency, InternalClientError
from neptune.new.internal.operation import (
    AssignArtifact,
    AddStrings,
    AssignBool,
    AssignDatetime,
    AssignFloat,
    AssignInt,
    AssignString,
    ClearFloatLog,
    ClearImageLog,
    ClearStringLog,
    ClearStringSet,
    ConfigFloatSeries,
    DeleteAttribute,
    DeleteFiles,
    LogFloats,
    LogImages,
    LogStrings,
    Operation,
    RemoveStrings,
    TrackFilesToArtifact,
    UploadFile,
    UploadFileContent,
    UploadFileSet,
)
from neptune.new.internal.operation_visitor import OperationVisitor
from neptune.new.internal.utils.paths import path_to_str

T = TypeVar("T")


class OperationsPreprocessor:

    def __init__(self):
        self._accumulators = dict()

    def process(self, operations: List[Operation]):
        for op in operations:
            path_str = path_to_str(op.path)
            self._accumulators.setdefault(path_str, _OperationsAccumulator(op.path)).visit(op)
        result = []
        for acc in self._accumulators.values():
            result.extend(acc.get_operations())
        return result

    def get_operations(self) -> List[Operation]:
        result = []
        for _, acc in sorted(self._accumulators.items()):
            result.extend(acc.get_operations())
        return result

    def get_errors(self) -> List[MetadataInconsistency]:
        result = []
        for _, acc in sorted(self._accumulators.items()):
            result.extend(acc.get_errors())
        return result


class _DataType(Enum):
    FLOAT = "Float"
    INT = "Int"
    BOOL = "Bool"
    STRING = "String"
    FILE = "File"
    DATETIME = "Datetime"
    FILE_SET = "File Set"
    FLOAT_SERIES = "Float Series"
    STRING_SERIES = "String Series"
    IMAGE_SERIES = "Image Series"
    STRING_SET = "String Set"
    ARTIFACT = "Artifact"


class _OperationsAccumulator(OperationVisitor[None]):

    def __init__(self, path: List[str]):
        self._path = path
        self._type = None
        self._delete_ops = []
        self._modify_ops = []
        self._config_ops = []
        self._errors = []

    def get_operations(self) -> List[Operation]:
        return self._delete_ops + self._modify_ops + self._config_ops

    def get_errors(self) -> List[MetadataInconsistency]:
        return self._errors

    def _process_modify_op(self,
                           expected_type: _DataType,
                           op: Operation,
                           modifier: Callable[[List[Operation], Operation], List[Operation]]) -> None:

        if self._type and self._type != expected_type:
            # This case should never happen since inconsistencies on data types are verified on user api.
            # So such operations should not appear in the queue without delete operation between them.
            # Still we want to support this case to avoid some unclear dependencies and assumptions.
            self._errors.append(MetadataInconsistency(
                "Cannot perform {} operation on {}: Attribute is not a {}".format(
                    op.__class__.__name__,
                    path_to_str(self._path),
                    expected_type.value)))
        else:
            self._type = expected_type
            self._modify_ops = modifier(self._modify_ops, op)

    def _process_config_op(self, expected_type: _DataType, op: Operation) -> None:

        if self._type and self._type != expected_type:
            # This case should never happen since inconsistencies on data types are verified on user api.
            # So such operations should not appear in the queue without delete operation between them.
            # Still we want to support this case to avoid some unclear dependencies and assumptions.
            self._errors.append(MetadataInconsistency(
                "Cannot perform {} operation on {}: Attribute is not a {}".format(
                    op.__class__.__name__,
                    path_to_str(self._path),
                    expected_type.value)))
        else:
            self._type = expected_type
            self._config_ops = [op]

    def visit_assign_float(self, op: AssignFloat) -> None:
        self._process_modify_op(_DataType.FLOAT, op, self._assign_modifier())

    def visit_assign_int(self, op: AssignInt) -> None:
        self._process_modify_op(_DataType.INT, op, self._assign_modifier())

    def visit_assign_bool(self, op: AssignBool) -> None:
        self._process_modify_op(_DataType.BOOL, op, self._assign_modifier())

    def visit_assign_string(self, op: AssignString) -> None:
        self._process_modify_op(_DataType.STRING, op, self._assign_modifier())

    def visit_assign_datetime(self, op: AssignDatetime) -> None:
        self._process_modify_op(_DataType.DATETIME, op, self._assign_modifier())

    def visit_upload_file(self, op: UploadFile) -> None:
        self._process_modify_op(_DataType.FILE, op, self._assign_modifier())

    def visit_upload_file_content(self, op: UploadFileContent) -> None:
        self._process_modify_op(_DataType.FILE, op, self._assign_modifier())

    def visit_assign_artifact(self, op: AssignArtifact) -> None:
        self._process_modify_op(_DataType.ARTIFACT, op, self._assign_modifier())

    def visit_upload_file_set(self, op: UploadFileSet) -> None:
        if op.reset:
            self._process_modify_op(_DataType.FILE_SET, op, self._assign_modifier())
        else:
            self._process_modify_op(_DataType.FILE_SET, op, self._add_modifier())

    def visit_log_floats(self, op: LogFloats) -> None:
        self._process_modify_op(
            _DataType.FLOAT_SERIES,
            op,
            self._log_modifier(LogFloats,
                               ClearFloatLog,
                               lambda op1, op2: LogFloats(op1.path, op1.values + op2.values)))

    def visit_log_strings(self, op: LogStrings) -> None:
        self._process_modify_op(
            _DataType.STRING_SERIES,
            op,
            self._log_modifier(LogStrings,
                               ClearStringLog,
                               lambda op1, op2: LogStrings(op1.path, op1.values + op2.values)))

    def visit_log_images(self, op: LogImages) -> None:
        self._process_modify_op(
            _DataType.IMAGE_SERIES,
            op,
            self._log_modifier(LogImages,
                               ClearImageLog,
                               lambda op1, op2: LogImages(op1.path, op1.values + op2.values)))

    def visit_clear_float_log(self, op: ClearFloatLog) -> None:
        self._process_modify_op(_DataType.FLOAT_SERIES, op, self._clear_modifier())

    def visit_clear_string_log(self, op: ClearStringLog) -> None:
        self._process_modify_op(_DataType.STRING_SERIES, op, self._clear_modifier())

    def visit_clear_image_log(self, op: ClearImageLog) -> None:
        self._process_modify_op(_DataType.IMAGE_SERIES, op, self._clear_modifier())

    def visit_add_strings(self, op: AddStrings) -> None:
        self._process_modify_op(_DataType.STRING_SET, op, self._add_modifier())

    def visit_clear_string_set(self, op: ClearStringSet) -> None:
        self._process_modify_op(_DataType.STRING_SET, op, self._clear_modifier())

    def visit_remove_strings(self, op: RemoveStrings) -> None:
        self._process_modify_op(_DataType.STRING_SET, op, self._remove_modifier())

    def visit_config_float_series(self, op: ConfigFloatSeries) -> None:
        self._process_config_op(_DataType.FLOAT_SERIES, op)

    def visit_delete_files(self, op: DeleteFiles) -> None:
        self._process_modify_op(_DataType.FILE_SET, op, self._add_modifier())

    def visit_delete_attribute(self, op: DeleteAttribute) -> None:
        if self._type:
            if self._delete_ops:
                # Keep existing delete operation and simply clear all modification operations after it
                self._modify_ops = []
                self._config_ops = []
                self._type = None
            else:
                # This case is tricky. There was no delete operation, but some modifications was performed.
                # We do not know if this attribute exists on server side and we do not want a delete op to fail.
                # So we need to send a single modification before delete to be sure a delete op is valid.
                self._delete_ops = [self._modify_ops[0], op]
                self._modify_ops = []
                self._config_ops = []
                self._type = None
        else:
            if self._delete_ops:
                # Do nothing if there already is a delete operation
                # and no other operations was performed after it
                return
            else:
                # If value has not been set locally yet and no delete operation was performed,
                # simply perform single delete operation.
                self._delete_ops.append(op)

    @staticmethod
    def _artifact_log_modifier(ops: List[TrackFilesToArtifact],
                               new_op: TrackFilesToArtifact) -> List[TrackFilesToArtifact]:
        if len(ops) == 0:
            return [new_op]

        # There should be exactly 1 operation, merge it with new_op
        assert len(ops) == 1
        op_old = ops[0]
        assert op_old.path == new_op.path
        assert op_old.project_id == new_op.project_id
        return [
            TrackFilesToArtifact(op_old.path, op_old.project_id, op_old.entries + new_op.entries)
        ]

    def visit_track_files_to_artifact(self, op: TrackFilesToArtifact) -> None:
        self._process_modify_op(
            _DataType.ARTIFACT,
            op,
            self._artifact_log_modifier
        )

    def visit_clear_artifact(self, op: ClearStringSet) -> None:
        self._process_modify_op(_DataType.ARTIFACT, op, self._clear_modifier())

    @staticmethod
    def _assign_modifier():
        return lambda ops, new_op: [new_op]

    @staticmethod
    def _clear_modifier():
        return lambda ops, new_op: [new_op]

    @staticmethod
    def _log_modifier(log_op_class: type,
                      clear_op_class: type,
                      log_combine: Callable[[T, T], T]):
        def modifier(ops, new_op):
            if len(ops) == 0:
                return [new_op]
            elif len(ops) == 1 and isinstance(ops[0], log_op_class):
                return [log_combine(ops[0], new_op)]
            elif len(ops) == 1 and isinstance(ops[0], clear_op_class):
                return [ops[0], new_op]
            elif len(ops) == 2:
                return [ops[0], log_combine(ops[1], new_op)]
            else:
                raise InternalClientError("Preprocessing operations failed: len(ops) == {}".format(len(ops)))
        return modifier

    @staticmethod
    def _add_modifier():
        # We do not optimize it on client side for now. It should not be often operation.
        return lambda ops, op: ops + [op]

    @staticmethod
    def _remove_modifier():
        # We do not optimize it on client side for now. It should not be often operation.
        return lambda ops, op: ops + [op]



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access

import unittest

from mock import patch, MagicMock

from neptune.new.internal.websockets.websocket_signals_background_job import WebsocketSignalsBackgroundJob


class TestClient(unittest.TestCase):

    @patch('neptune.new.internal.websockets.websocket_signals_background_job.process_killer')
    def test_listener_stop(self, process_killer):
        # given
        run, ws = MagicMock(), MagicMock()
        listener = WebsocketSignalsBackgroundJob._ListenerThread(run, ws)

        ws.recv.return_value = '{"type": "neptune/stop", "body": {"seconds": 5}}'

        # when
        listener.work()

        # then
        self.assertEqual(0, run.__setitem__.call_count)
        run.stop.assert_called_once_with(seconds=5)
        process_killer.kill_me.assert_called_once_with()

    @patch('neptune.new.internal.websockets.websocket_signals_background_job.process_killer')
    def test_listener_abort(self, process_killer):
        # given
        run, ws = MagicMock(), MagicMock()
        listener = WebsocketSignalsBackgroundJob._ListenerThread(run, ws)

        ws.recv.return_value = '{"type": "neptune/abort", "body": {"seconds": 5}}'

        # when
        listener.work()

        # then
        run.__setitem__.assert_called_once_with("sys/failed", True)
        run.stop.assert_called_once_with(seconds=5)
        process_killer.kill_me.assert_called_once_with()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest

from neptune.new.exceptions import MetadataInconsistency
from neptune.new.internal.run_structure import RunStructure


class TestRunStructure(unittest.TestCase):

    def test_get_none(self):
        exp = RunStructure[int, dict]()
        self.assertEqual(exp.get(["some", "path", "val"]), None)

    def test_get_nested_variable_fails(self):
        exp = RunStructure[int, dict]()
        exp.set(["some", "path", "val"], 3)
        with self.assertRaises(MetadataInconsistency):
            exp.get(["some", "path", "val", "nested"])
        with self.assertRaises(MetadataInconsistency):
            exp.get(["some", "path", "val", "nested", "nested"])

    def test_get_ns(self):
        exp = RunStructure[int, dict]()
        exp.set(["some", "path", "val"], 3)
        self.assertEqual(exp.get(["some", "path"]), {"val": 3})

    def test_set(self):
        exp = RunStructure[int, dict]()
        exp.set(["some", "path", "val"], 3)
        self.assertEqual(exp.get(["some", "path", "val"]), 3)

    def test_set_nested_variable_fails(self):
        exp = RunStructure[int, dict]()
        exp.set(["some", "path", "val"], 3)
        with self.assertRaises(MetadataInconsistency):
            exp.set(["some", "path", "val", "nested"], 3)
        with self.assertRaises(MetadataInconsistency):
            exp.set(["some", "path", "val", "nested", "nested"], 3)

    def test_set_ns_collision(self):
        exp = RunStructure[int, dict]()
        exp.set(["some", "path", "val"], 3)
        with self.assertRaises(MetadataInconsistency):
            exp.set(["some", "path"], 5)

    def test_pop(self):
        exp = RunStructure[int, dict]()
        exp.set(["some", "path", "val1"], 3)
        exp.set(["some", "path", "val2"], 5)
        exp.pop(["some", "path", "val2"])
        self.assertEqual(exp.get(["some", "path", "val1"]), 3)
        self.assertEqual(exp.get(["some", "path", "val2"]), None)
        self.assertTrue("some" in exp.get_structure() and "path" in exp.get_structure()["some"])

    def test_pop_whole_ns(self):
        exp = RunStructure[int, dict]()
        exp.set(["some", "path", "val"], 3)
        exp.pop(["some", "path", "val"])
        self.assertEqual(exp.get(["some", "path", "val"]), None)
        self.assertFalse("some" in exp.get_structure())

    def test_pop_not_found(self):
        exp = RunStructure[int, dict]()
        with self.assertRaises(MetadataInconsistency):
            exp.pop(["some", "path"])

    def test_pop_ns_fail(self):
        exp = RunStructure[int, dict]()
        exp.set(["some", "path", "val1"], 3)
        with self.assertRaises(MetadataInconsistency):
            exp.pop(["some", "path"])



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import os
import unittest

from neptune.new import ANONYMOUS, init
from neptune.new.envs import API_TOKEN_ENV_NAME, PROJECT_ENV_NAME
from neptune.new.integrations.python_logger import NeptuneHandler


class TestLogHandler(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        os.environ[PROJECT_ENV_NAME] = "organization/project"
        os.environ[API_TOKEN_ENV_NAME] = ANONYMOUS

    def _log_messages(self, logger):
        logger.error("error message")
        logger.debug("debug message")
        logger.warning("test %s", "message")

    def test_no_run(self):
        with self.assertRaises(TypeError):
            NeptuneHandler(run="PET-1")

    def test_default_attribute(self):
        exp = init(mode="debug", flush_period=0.5)
        handler = NeptuneHandler(run=exp)
        logger = logging.getLogger()
        logger.addHandler(handler)

        self._log_messages(logger)

        log_entries = list(exp['monitoring']['python_logger'].fetch_values().value)
        self.assertListEqual(log_entries, ["error message", "test message"])

    def test_custom_monitoring_namespace(self):
        exp = init(mode="debug", flush_period=0.5, monitoring_namespace="watching")
        handler = NeptuneHandler(run=exp)
        logger = logging.getLogger()
        logger.addHandler(handler)

        self._log_messages(logger)

        log_entries = list(exp['watching']['python_logger'].fetch_values().value)
        self.assertListEqual(log_entries, ["error message", "test message"])

    def test_custom_target_attribute(self):
        exp = init(mode="debug", flush_period=0.5)
        handler = NeptuneHandler(run=exp, path="logging/my/logger")
        logger = logging.getLogger()
        logger.addHandler(handler)

        self._log_messages(logger)

        log_entries = list(exp['logging']['my']['logger'].fetch_values().value)
        self.assertListEqual(log_entries, ["error message", "test message"])
        self.assertNotIn('python_logger', exp.get_structure()['monitoring'])

    def test_custom_level(self):
        exp = init(mode="debug", flush_period=0.5)
        handler = NeptuneHandler(run=exp, level=logging.ERROR)
        logger = logging.getLogger()
        logger.addHandler(handler)

        self._log_messages(logger)

        log_entries = list(exp['monitoring']['python_logger'].fetch_values().value)
        self.assertListEqual(log_entries, ["error message"])

    def test_formatter_works(self):
        exp = init(mode="debug", flush_period=0.5)
        handler = NeptuneHandler(run=exp)
        handler.setFormatter(logging.Formatter("%(levelname)s|%(name)s: %(message)s"))
        logger = logging.getLogger()
        logger.addHandler(handler)

        self._log_messages(logger)

        log_entries = list(exp['monitoring']['python_logger'].fetch_values().value)
        self.assertListEqual(log_entries, ["ERROR|root: error message", "WARNING|root: test message"])

    def test_log_level_works(self):
        exp = init(mode="debug", flush_period=0.5)
        handler = NeptuneHandler(run=exp)
        logger = logging.getLogger()
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)

        self._log_messages(logger)

        log_entries = list(exp['monitoring']['python_logger'].fetch_values().value)
        self.assertListEqual(log_entries, ["error message", "debug message", "test message"])

        self._log_messages(logger)

    def test_log_level_works_with_level(self):
        exp = init(mode="debug", flush_period=0.5)
        handler = NeptuneHandler(run=exp, level=logging.WARNING)
        logger = logging.getLogger()
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)

        self._log_messages(logger)

        log_entries = list(exp['monitoring']['python_logger'].fetch_values().value)
        self.assertListEqual(log_entries, ["error message", "test message"])

        handler.setLevel(logging.ERROR)

        self._log_messages(logger)

        log_entries = list(exp['monitoring']['python_logger'].fetch_values().value)
        self.assertListEqual(log_entries, ["error message", "test message", "error message"])



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from datetime import datetime
from typing import TypeVar, TYPE_CHECKING

from neptune.new.types.atoms.atom import Atom

if TYPE_CHECKING:
    from neptune.new.types.value_visitor import ValueVisitor

Ret = TypeVar('Ret')


class Datetime(Atom):

    def __init__(self, value: datetime):
        self.value = value.replace(microsecond=1000*int(value.microsecond/1000))

    def accept(self, visitor: 'ValueVisitor[Ret]') -> Ret:
        return visitor.visit_datetime(self)

    def __str__(self):
        return "Datetime({})".format(str(self.value))



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import typing
import pathlib
import hashlib


def sha1(fname: typing.Union[str, pathlib.Path], block_size: int = 2 ** 16) -> str:
    sha1sum = hashlib.sha1()

    with open(fname, 'rb') as source:
        block = source.read(block_size)

        while len(block) != 0:
            sha1sum.update(block)
            block = source.read(block_size)

    return str(sha1sum.hexdigest())



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
from typing import Optional

from neptune.management.exceptions import InvalidProjectName, ConflictingWorkspaceName, MissingWorkspaceName
from neptune.patterns import PROJECT_QUALIFIED_NAME_PATTERN


def normalize_project_name(name: str, workspace: Optional[str] = None):
    project_spec = re.search(PROJECT_QUALIFIED_NAME_PATTERN, name)

    if not project_spec:
        raise InvalidProjectName(name=name)

    extracted_workspace, extracted_name = project_spec['workspace'], project_spec['project']

    if not workspace and not extracted_workspace:
        raise MissingWorkspaceName(name=name)

    if workspace and extracted_workspace and workspace != extracted_workspace:
        raise ConflictingWorkspaceName(name=name, workspace=workspace)

    return f'{extracted_workspace or workspace}/{extracted_name}'
